/**
 * Main executable class of Ipauth. The executable JAR can be run with the
 * following command:
 * 
 * java -jar ipauth.jar [target] [options]
 * 
 * For more information about the usage, run JAR with -h or --help option.
 * 
 * @author Chiache Tsai
 */

package org.netserv.ipauth;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import java.io.BufferedReader;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Vector;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
import org.bouncycastle.asn1.x509.X509CertificateStructure;
import org.bouncycastle.openssl.PEMReader;
import org.netserv.ipauth.core.*;
import org.netserv.ipauth.impl.*;
import org.netserv.ipauth.service.RaServer;
import org.netserv.ipauth.util.*;
import org.netserv.ipauth.ui.*;

public class IpAuthMain {

  public static void printUsage() {
    System.err.println("USAGE: java -jar ipauth.jar target [args]");
    System.err.println("For more detail info, use -h or --help argument");
  }

  public static void main(String args[]) {
    if (args.length == 0) {
      printUsage();
      return;   
    }
    
    String target = args[0];
    System.arraycopy(args, 1, args, 0, args.length - 1);
    
    if (target.equalsIgnoreCase("sign")) {
      doSign(args);
    }
    else if (target.equalsIgnoreCase("verify")) {
      doVerify(args);
    }
    else if (target.equalsIgnoreCase("prove")) {
      doProve(args);
    }
    else if (target.equalsIgnoreCase("socket")) {
      doSocket(args);
    }
    else if (target.equalsIgnoreCase("client")) {
      doClient(args);
    }
    else if (target.equalsIgnoreCase("ra")) {
      doRA(args);
    }
    else {
      printUsage();
      return;     	
    }
  }

  public static void printSignUsage() {
    System.err.println("[USAGE] java -jar ipauth.jar sign [--help|-h] " +
                       "[--cert|-c file] [--key|-k file] " +
                       "[--impl|-i implementation] [--alg|-a algorithm]");
  }

  /**
   * @param args
   */
  public static void doSign(String args[]) {
    LongOpt[] longopts = new LongOpt[5];
    longopts[0] = new LongOpt("help", LongOpt.NO_ARGUMENT,       null, 'h');
    longopts[1] = new LongOpt("cert", LongOpt.REQUIRED_ARGUMENT, null, 'c');
    longopts[2] = new LongOpt("key" , LongOpt.REQUIRED_ARGUMENT, null, 'k');
    longopts[3] = new LongOpt("impl", LongOpt.REQUIRED_ARGUMENT, null, 'i');
    longopts[4] = new LongOpt("alg" , LongOpt.REQUIRED_ARGUMENT, null, 'a');
    Getopt g = new Getopt("ipauth sign", args, "hc:k:i:a:", longopts);
    int c;

    String certFile  = null;
    String keyFile   = null;
    String impl      = null;
    String algorithm = null;
    while ((c = g.getopt()) != -1) {
      switch (c) {
      case 'h': printSignUsage(); return;
      case 'c': certFile  = g.getOptarg(); break;
      case 'k': keyFile   = g.getOptarg(); break;
      case 'i': impl      = g.getOptarg(); break;
      case 'a': algorithm = g.getOptarg(); break;
      default:  printSignUsage(); return;
      }
    }

    if (certFile == null || keyFile   == null || 
        impl     == null || algorithm == null) {
      printSignUsage();
      return;
    }

    CryptoHelper.installBC();
    
    X509Certificate[] certChain = FileHelper.readCertificates(certFile);
    PrivateKey key = FileHelper.readKeyPair(keyFile, null).getPrivate();

    if (certChain == null || key == null)
      return;
    
    IpAuthSigner signer = null;
    if (impl.equalsIgnoreCase("jar")) {
      signer = new IpAuthSigner(new JarSigner(key, algorithm, "BC"));
    }
    else if (impl.equalsIgnoreCase("cms")) {
      signer = new IpAuthSigner(new CmsSigner(key, algorithm, "BC"));
    }
    else {
      System.err.println("Signer Implementation Not Supported: " + impl);
      return;
    }

    for (X509Certificate cert: certChain) {
      signer.addCaPub(cert);
    }
    
    try {
      signer.signStream(new FileInputStream(FileDescriptor.in), 
                        new FileOutputStream(FileDescriptor.out));
    } catch (IOException e) { 
      e.printStackTrace();
      return; 
    } catch (CertificateException e) {
	  e.printStackTrace();
	  return;
	}
  }

  /**
   *
   */
  public static void printVerifyUsage() {
    System.err.println("[USAGE] java -jar bcclient.jar VERIFY [--help|-h] " +
           "[--trust|-t file] [--dist|-d file]");
  }

  /**
   *
   * @param args
   */
  public static void doVerify(String args[]) {
    LongOpt[] longopts = new LongOpt[5];
    longopts[0] = new LongOpt("help", LongOpt.NO_ARGUMENT,   null, 'h');
    longopts[1] = new LongOpt("trust",LongOpt.REQUIRED_ARGUMENT, null, 't');
    longopts[2] = new LongOpt("dist", LongOpt.REQUIRED_ARGUMENT, null, 's');
    longopts[3] = new LongOpt("impl", LongOpt.REQUIRED_ARGUMENT, null, 'i');
    longopts[4] = new LongOpt("depth",LongOpt.REQUIRED_ARGUMENT, null, 'd');
    Getopt g = new Getopt("ipauth verify", args, "ht:d:i:", longopts);
    int c;

    String trustFile = null;
    String distFile  = null;
    String impl      = null;
    int    depth     = 0;
    while ((c = g.getopt()) != -1) {
      switch (c) {
      case 'h': printVerifyUsage(); return;
      case 't': trustFile = g.getOptarg(); break;
      case 's': distFile  = g.getOptarg(); break;
      case 'i': impl  = g.getOptarg(); break;
      case 'd': depth = Integer.parseInt(g.getOptarg()); break;
      }
    }

    IpAuthVerifier verifier;
    if (impl.equalsIgnoreCase("crl")) {
      verifier = new IpAuthVerifier(new CrlVerifier(), depth);
    }
    else {
      System.err.println("Verifier Implementation Not Supported: " + impl);
      return;
    }
    
    if (trustFile != null) {
      X509Certificate[] trustArray = FileHelper.readCertificates(trustFile);
      verifier.registerTrustedCA(trustArray);
    }

    if (distFile != null) {
      Vector<URL> dists = new Vector<URL>();
      try {
      BufferedReader reader = new BufferedReader(new FileReader(distFile));
      String line = null;
      while((line = reader.readLine()) != null)
        dists.add(new URL(line));
      } catch (IOException ex) {}

      verifier.registerTrustedDists(dists.toArray(new URL[dists.size()]));
    }

    PEMReader reader = new PEMReader(new InputStreamReader(System.in));
    Object o = null;
    Vector<X509Certificate> certs = new Vector<X509Certificate>();
    try {
      while ((o = reader.readObject()) != null) {
      if (o instanceof X509Certificate)
        certs.add((X509Certificate)o);
      }
    } catch (IOException ex) {}

    try {
	  verifier.verify(certs.toArray(new X509Certificate[certs.size()]));
	} catch (CertificateException e) {
	  System.err.println("Certificate verification failed");
	  return;
	}
  }

  /**
  *
  */
 public static void printProveUsage() {
   System.err.println("[USAGE] java -jar bcclient.jar PROVER [--help|-h] " +
                      "[--config|-c file]");
 }

 /**
  *
  * @param args
  */
 public static void doProve(String args[]) {
/*
   LongOpt[] longopts = new LongOpt[2];
   longopts[0] = new LongOpt("help",   LongOpt.NO_ARGUMENT,       null, 'h');
   longopts[1] = new LongOpt("config", LongOpt.REQUIRED_ARGUMENT, null, 'c');
   Getopt g = new Getopt("ipauth PROVER", args, "hc:", longopts);
   int c;

   String configFile = null;
   while ((c = g.getopt()) != -1) {
     switch (c) {
     case 'h': printProveUsage(); return;
     case 'c': configFile = g.getOptarg(); break;
     default:
     }
   }

   try {
     IpAuthProver server = new IpAuthProver(configFile);
     server.start();
   } catch (Exception ex) {
     ex.printStackTrace();
   }
 */
 }
  

 /**
  *
  */
 public static void printSocketUsage() {
   System.err.println("[USAGE] java -jar bcclient.jar CRL [--help|-h] " +
              "[--listen|-l port] [--addr|-a addr] [--port|-p port] " +
              "[--verify|-v] [--level level] [--trust file] [--dist file] " +
              "[--sign|-s] [--cert file] [--key file]");
 }
 
 /**
 *
 * @param args
 */
public static void doSocket(String args[]) {
	/*
  LongOpt[] longopts = new LongOpt[11];
  longopts[0]  = new LongOpt("help",   LongOpt.NO_ARGUMENT,       null, 'h');
  longopts[1]  = new LongOpt("listen", LongOpt.REQUIRED_ARGUMENT, null, 'l');
  longopts[2]  = new LongOpt("addr",   LongOpt.REQUIRED_ARGUMENT, null, 'a');
  longopts[3]  = new LongOpt("port",   LongOpt.REQUIRED_ARGUMENT, null, 'p');
  longopts[4]  = new LongOpt("verify", LongOpt.NO_ARGUMENT,       null, 'v');
  longopts[5]  = new LongOpt("level",  LongOpt.REQUIRED_ARGUMENT, null, '1');
  longopts[6]  = new LongOpt("trust",  LongOpt.REQUIRED_ARGUMENT, null, '2');
  longopts[7]  = new LongOpt("dist",   LongOpt.REQUIRED_ARGUMENT, null, '3');
  longopts[8]  = new LongOpt("sign",   LongOpt.NO_ARGUMENT,       null, 's');
  longopts[9]  = new LongOpt("cert",   LongOpt.REQUIRED_ARGUMENT, null, '4');
  longopts[10] = new LongOpt("key",    LongOpt.REQUIRED_ARGUMENT, null, '5');
  Getopt g = new Getopt("ipauth SOCKET", args, "hl:a:p:vs1:2:3:4:5:", longopts);
  int c;

  boolean listen = false;
  int    bind    = 0;
  String addr    = null;
  int    port    = 0;
  boolean verify = false;
  int level = 0;
  String  trustFile  = null;
  String  distFile   = null;
  boolean sign   = false;
  String  certFile = null;
  String  keyFile  = null;
  while ((c = g.getopt()) != -1) {
    switch (c) {
      case 'h': printSocketUsage(); return;
      case 'l': listen = true; bind = Integer.parseInt(g.getOptarg()); break;
      case 'a': addr = g.getOptarg(); break;
      case 'p': port = Integer.parseInt(g.getOptarg()); break;
      case 'v': verify = true; break;
      case '1': level = Integer.parseInt(g.getOptarg()); break;
      case '2': trustFile = g.getOptarg(); break;
      case '3': distFile = g.getOptarg(); break;
      case 's': sign = true; break;
      case '4': certFile = g.getOptarg(); break;
      case '5': keyFile = g.getOptarg(); break;
    }
  }

  IpAuthVerifierImpl verifier = null;
  if (verify) {
    verifier = new CrlVerifier(level);
    if (trustFile != null) {
      Vector<X509Certificate> trusts = new Vector<X509Certificate>();
      X509Certificate[] trustArray = FileHelper.readCertificate(trustFile);
      for (X509Certificate trust: trustArray) {
      trusts.add(trust);
      }
      verifier.registerTrustedCA(trusts);
    }

    if (distFile != null) {
      Vector<URL> dists = new Vector<URL>();
      try {
      BufferedReader reader = new BufferedReader(new FileReader(distFile));
      String line = null;
      while((line = reader.readLine()) != null)
        dists.add(new URL(line));
      } catch (IOException ex) {}

      verifier.registerTrustedDists(dists);
    }
  }

  IpAuthSignerImpl signer = null;
  if (sign) {
    signer = null;
    X509CertificateStructure[] certStructs =
                              FileHelper.readCertificateStructure(certFile);
    KeyPair key = FileHelper.readKeyPair(keyFile, null);
    signer = new IpAuthSignerImpl(certStructs, key.getPrivate(),
      new AlgorithmIdentifier(PKCSObjectIdentifiers.sha256WithRSAEncryption));
  }

  SignedSocket socket;
  if (listen) {
    try {
      ServerSocket server = new ServerSocket(bind, 0, InetAddress.getByName("0.0.0.0"));
      socket = new SignedSocket(server.accept(), verifier, signer);
      
    } catch (IOException ex) {
      ex.printStackTrace();
      socket = null;
    }
  } else {
    try {
      socket = new SignedSocket(new Socket(addr, port), verifier, signer);
    } catch (Exception ex) {
      ex.printStackTrace();
      socket = null;
    }
  }
  if (socket == null)
    return;

  try {
    final InputStream  is = socket.getInputStream();
    final OutputStream os = socket.getOutputStream();

    new Thread(new Runnable() {
      byte[] buffer = new byte[4096];

      public void run() {
        try {
          while (true) {
            int read = is.read(buffer);
            if (read < 0)
              break;
            System.out.write(buffer, 0, read);
            System.out.write('\n');
          }
        } catch (IOException ex) {}
      }

    }).start();

    new Thread(new Runnable() {
      byte[] buffer = new byte[4096];

      public void run() {
        try {
          while (true) {
            int read = System.in.read(buffer);
            if (read < 0)
              break;
            os.write(buffer, 0, read);
            os.flush();
          }
        } catch (IOException ex) {}
      }

    }).start();

  } catch (Exception ex) {
    ex.printStackTrace();
  }
  */
}
 
  /**
   *
   */
  public static void printClientUsage() {
    System.err.println("[USAGE] java -jar bcclient.jar CLIENT [--help|-h] " +
                       "[--config|-c file] [--gui|-g]");
  }

  /**
   *
   * @param args
   */
  public static void doClient(String args[]) {
    LongOpt[] longopts = new LongOpt[3];
    longopts[0] = new LongOpt("help",   LongOpt.NO_ARGUMENT,       null, 'h');
    longopts[1] = new LongOpt("config", LongOpt.REQUIRED_ARGUMENT, null, 'c');
    longopts[2] = new LongOpt("gui",    LongOpt.NO_ARGUMENT,       null, 'g');
    Getopt g = new Getopt("ipauth Client", args, "hc:g", longopts);
    int c;

    String configFile = null;
    boolean gui = false;
    while ((c = g.getopt()) != -1) {
      switch (c) {
      case 'h': printClientUsage(); return;
      case 'c': configFile = g.getOptarg(); break;
      case 'g': gui = true;
      }
    }

    try {
      IpAuthClient client = new IpAuthClient(configFile);

      if (gui)
        new IpAuthClientGUI(client).setVisible(true);
      else
        client.request();
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }



  /**
   *
   */
  public static void printRAUsage() {
    System.err.println("[USAGE] java -jar bcclient.jar RA [--help|-h] " +
                       "[--config|-c file]");
  }

  /**
   *
   * @param args
   */
  public static void doRA(String args[]) {
    LongOpt[] longopts = new LongOpt[2];
    longopts[0] = new LongOpt("help",   LongOpt.NO_ARGUMENT,     null, 'h');
    longopts[1] = new LongOpt("config", LongOpt.REQUIRED_ARGUMENT, null, 'c');
    Getopt g = new Getopt("ipauth RA", args, "hc:", longopts);
    int c;

    String configFile = null;
    while ((c = g.getopt()) != -1) {
      switch (c) {
        case 'h': printRAUsage(); return;
        case 'c': configFile = g.getOptarg(); break;
      }
    }

    try {
      RaServer server = new RaServer(configFile);
      server.start();
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }

}

