package org.netserv.ipauth.core;

import java.math.BigInteger;
import java.net.URL;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import java.util.Vector;
import org.bouncycastle.asn1.ASN1EncodableVector;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.DEREncodable;
import org.bouncycastle.asn1.DERInteger;
import org.bouncycastle.asn1.DERObject;
import org.bouncycastle.asn1.DERSequence;

/** 
 * The individual result of certificate checking. class <code>CertStatus</code>
 * contains a certificate set validation algorithm, which is general enough
 * to work on all certificate verification protocol.
 * 
 * ASN.1 structure of <code>CertStatus</code>:
 * <pre>
 *  CertStatus ::= SEQUENCE {
 *    serialNumber INTEGER,
 *    validDepth  INTEGER,
 *  }
 * </pre>
 * 
 * @see org.netserv.ipauth.core.IpAuthVerifier
 * @see org.netserv.ipauth.core.IpAuthVerifierImpl
 * 
 * @author Chiache
 */

public class CertStatus implements DEREncodable
{
  public static final int DEPTH_CHECKED_BAD  = -127;
  public static final int DEPTH_TRUSTED      =  0;
  public static final int DEPTH_CHECKED_GOOD =  127;
  public static final int DEPTH_UNCHECKED    =  128;

  private X509Certificate certificate;
  private BigInteger      serialNumber;
  private int             validDepth;

  private Vector<URL> distPoints = new Vector<URL>();

  private CertStatus parent = null;
  private Vector<CertStatus> children = new Vector<CertStatus>();

  public CertStatus() {
    validDepth   = DEPTH_UNCHECKED;
  }

  public CertStatus(X509Certificate cert) {
    certificate  = cert;
    serialNumber = cert.getSerialNumber();
    validDepth   = DEPTH_UNCHECKED;
  }

  public CertStatus(BigInteger serial, int depth) {
    serialNumber = serial;
    validDepth   = depth;
  }

  public CertStatus(ASN1Sequence seq) {
    Enumeration e = seq.getObjects();

    serialNumber = DERInteger.getInstance(e.nextElement()).getPositiveValue();
    validDepth   = DERInteger.getInstance(e.nextElement()).getValue().intValue();
  }

  public void setCertificate(X509Certificate cert) {
    certificate = cert;
  }

  public X509Certificate getCertificate() {
    return certificate; 
  }

  public void setSerial(BigInteger serial) {
    serialNumber = serial;
  }

  public BigInteger getSerial() {
    return serialNumber;
  }

  public void setValidDepth(int depth) {
    if (certificate != null)

    if (depth >= validDepth)
      return;

    if (depth < 0) {
      for (CertStatus cp: children) {
        if (cp.getValidDepth() >= 0)
          cp.setValidDepth(DEPTH_CHECKED_BAD);
      }
      validDepth = DEPTH_CHECKED_BAD;
    }
    else {
      for (CertStatus cp: children) {
        if (cp.getValidDepth() != DEPTH_UNCHECKED &&
          cp.getValidDepth() == validDepth)
          cp.setValidDepth(depth);
        else
          cp.setValidDepth(depth + 1);
      }
      validDepth = depth;
    }
  }

  public int getValidDepth() {
    return validDepth;
  }

  public void addDistPoint(URL distPoint) {
    distPoints.add(distPoint);
  }

  public void addAllDistPoint(Vector<URL> distPoint) {
    distPoints.addAll(distPoint);
  }

  public Vector<URL> getDistPoint() {
    return distPoints;
  }

  public void setParent(CertStatus cert) {
    parent = cert;
  }

  public CertStatus getParent() {
    return parent;
  }

  public void addChild(CertStatus cert) {
    children.add(cert);
  }

  public Vector<CertStatus> getChildren() {
    return children;
  }

  public static CertStatus getInstance(Object obj)
  {
    if (obj instanceof CertStatus)
    {
      return (CertStatus)obj;
    }
    else if (obj instanceof ASN1Sequence)
    {
      return new CertStatus((ASN1Sequence)obj);
    }

    throw new IllegalArgumentException("unknown object in factory");
  }

  public DERObject getDERObject() {
    ASN1EncodableVector  v = new ASN1EncodableVector();

    v.add(new DERInteger(serialNumber));
    v.add(new DERInteger(validDepth));

    return new DERSequence(v);
  }
}
