package org.netserv.ipauth.core;

import java.io.IOException;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateException;
import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpServer;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.DEREncodable;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x509.X509CertificateStructure;
import com.novosec.pkix.asn1.cmp.PKIBody;
import com.novosec.pkix.asn1.cmp.PKIMessage;
import com.novosec.pkix.asn1.cmp.RevReqContent;
import com.novosec.pkix.asn1.crmf.CertReqMessages;
import com.novosec.pkix.asn1.crmf.CertTemplate;
import com.sun.net.httpserver.HttpHandler;
import java.util.Vector;

import org.netserv.ipauth.util.CmpHelper;
import org.netserv.ipauth.util.CryptoHelper;

/**
 * Class <code>IpAuthProver</code> represents provers in IpAuth frameworks. 
 * <code>IpAuthProver</code> may be used in the client side or in the server
 * side. In the client side, this class provides a method <code>obtain()</code>
 * to obtain IP addresses bound to the public key. In the server side, this
 * class provides a method <code>createServer</code> to create a prover daemon
 * working as a http server.
 * 
 * @see org.netserv.ipauth.core.IpAuthProverImpl
 * 
 * @author Chiache Tsai
 */

public class IpAuthProver {

  private IpAuthProverImpl prover;

  private PrivateKey privateKey;
  private Vector<X509CertificateStructure> caPubs;

  public IpAuthProver(PrivateKey key, IpAuthProverImpl prover) throws IOException {
    CryptoHelper.installBC();
    privateKey = key;
  }

  public void addCaPubs(X509CertificateStructure caPub) {
    caPubs.add(caPub);
  }

  public byte[] obtain(PublicKey key, String device) {
    return prover.obtain(key, device);
  }
  
  public HttpServer createServer(InetSocketAddress address, String distPoint)
          throws IOException {
    HttpServer server = HttpServer.create();
    server.bind(address, 0);
    server.createContext(distPoint, new HttpHandler() {
      public void handle(HttpExchange exchange) throws IOException {
        handlerMethod(exchange);
      }
    });
    return server;
  }

  protected void handlerMethod(HttpExchange he) throws IOException {
    if (he.getRequestMethod().equals("POST")) {

      ASN1InputStream  asn1is = new ASN1InputStream(he.getRequestBody());

      PKIMessage pkireq;
      try {
        pkireq = PKIMessage.getInstance(asn1is.readObject());
        if (pkireq == null)
          throw new CertificateException("no PKI message");

        PKIBody body = pkireq.getBody();
        DEREncodable certReqs = null;
        switch (body.getTagNo()) {
          case 0:  certReqs = body.getIr();  break;
          case 2:  certReqs = body.getCr();  break;
          case 7:  certReqs = body.getKur(); break;
          case 9:  certReqs = body.getKrr(); break;
          case 11: certReqs = body.getRp();  break;
          case 13: certReqs = body.getCcr(); break;
        }

        if (certReqs == null)
          throw new Exception();

        CertTemplate certTemp = null;
        if (certReqs instanceof CertReqMessages) {
          if (((CertReqMessages)certReqs).getCertReqMsg(1) != null)
            throw new Exception();
          certTemp = ((CertReqMessages)certReqs).getCertReqMsg(0).getCertReq().getCertTemplate();
        }
        else if (certReqs instanceof RevReqContent) {
          if (((RevReqContent)certReqs).getRevDetails(1) != null)
            throw new Exception();
          certTemp = ((RevReqContent)certReqs).getRevDetails(0).getCertDetails();
        }

        if (certTemp == null)
          throw new Exception();

        String name = CryptoHelper.getCommonNames(certTemp.getSubject())[0];
        PublicKey key = CryptoHelper.SubjectPublicKeyInfoToPublicKey(certTemp.getPublicKey());

        System.err.println("[PROVER] Get a request for IP " + name + " from " +
                           he.getRemoteAddress().getAddress().getHostAddress());
        if (!InetAddress.getByName(name).equals(he.getRemoteAddress().getAddress()))
          throw new Exception();

        byte[] address = he.getRemoteAddress().getAddress().getAddress();
        if (!prover.prove(address, key))
          throw new Exception();

        pkireq = CmpHelper.protectPKIMessageWithCert
                (pkireq, caPubs.firstElement(), privateKey,
                 PKCSObjectIdentifiers.sha256WithRSAEncryption.getId());
        for(X509CertificateStructure caPub: caPubs)
          pkireq.addExtraCert(caPub);

        he.sendResponseHeaders(200, 0);
        OutputStream responseBody = he.getResponseBody();
        responseBody.write(pkireq.getDERObject().getEncoded());
        responseBody.close();
      } catch (Exception ex) {
        ex.printStackTrace();
        he.sendResponseHeaders(405, 0);
        OutputStream responseBody = he.getResponseBody();
        responseBody.write("Invalid request: ".getBytes());
        responseBody.write(ex.getMessage().getBytes());
        responseBody.close();
        return;
      }
    }
    else {
      System.err.println("[PROVER] A connection is denied: using 'POST' method");
      he.sendResponseHeaders(405, 0);
      OutputStream responseBody = he.getResponseBody();
      responseBody.write("You can only use 'POST' method".getBytes());
      responseBody.close();
    }
  }
}
