package org.netserv.ipauth.core;

import java.math.BigInteger;
import java.net.URL;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Vector;

/**
 * IpAuthVerifier is the generic interface of Verifier in IpAuth Framework.
 * A Verifier is a role in the PKI system, which uses some certificate 
 * verification mechanism to determine whether the certificate is good or
 * not. The behavior of certificate verification is defined by subclasses of 
 * IpAuthVerifierImpl. IpAuthVerifier provides a verification algorithm.
 * 
 * @see org.netserv.ipsuth.core.IpAuthVerifierImpl
 * 
 * @author Chiache Tsai
 */

public final class IpAuthVerifier {

  private IpAuthVerifierImpl verifier;
  private int maxDepth = 0;

  public IpAuthVerifier(IpAuthVerifierImpl verifier, int depth) {
    this.verifier = verifier;
    this.maxDepth = depth;
  }

  public void registerTrustedCA(X509Certificate[] caCerts) {
    Vector<X509Certificate> caCertsV = new Vector<X509Certificate>();
    caCertsV.copyInto(caCerts);
    verifier.registerTrustedCA(caCertsV);
  }

  public void registerTrustedDists(URL[] dists) {
    Vector<URL> distsV = new Vector<URL>();
    distsV.copyInto(dists);
    verifier.registerTrustedDists(distsV);
  }

  public void verify(X509Certificate[] certs) throws CertificateException {

    CertStatus[] status = verifier.checkStatus(certs);

    for (X509Certificate cert: certs) {
      BigInteger serial = cert.getSerialNumber();
      boolean found = false;
      for (CertStatus stat: status) {
        if (stat.getSerial().equals(serial)) {
          if (stat.getValidDepth() > maxDepth ||
              stat.getValidDepth() <= CertStatus.DEPTH_CHECKED_BAD)
            throw new CertificateException
                    (cert.getSubjectDN() + " depth=" + stat.getValidDepth());
          found = true;
          break;
        }
      }
      if (!found)
        throw new CertificateException
                (cert.getSubjectDN() + " not checked");
    }
  }

}
