package org.netserv.ipauth.core;

import org.netserv.ipauth.util.CryptoHelper;
import java.net.URL;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Hashtable;
import java.util.Vector;

/**
 *
 * @author Chiache Tsai
 */

public abstract class IpAuthVerifierImpl  {

  protected Vector<URL> trustedDists = new Vector<URL>();
  protected Hashtable<X509Certificate, CertStatus> certStore
                = new Hashtable<X509Certificate, CertStatus>();

  public IpAuthVerifierImpl() {
    CryptoHelper.installBC();
  }

  public IpAuthVerifierImpl(int acceptDepth) {
    if (acceptDepth < 0 || acceptDepth > CertStatus.DEPTH_CHECKED_GOOD)
      throw new IllegalArgumentException();
  }

  public void registerTrustedCA(Vector<X509Certificate> caCerts) {
    for (X509Certificate cert: caCerts) {
      CertStatus profile = new CertStatus(cert);
      profile.setValidDepth(CertStatus.DEPTH_TRUSTED);
      insert(cert, profile);
    }
  }

  public void registerTrustedDists(Vector<URL> dists) {
    trustedDists.addAll(dists);
  }

  protected abstract Vector<URL> retrieveDistPoint(X509Certificate cert);

  protected abstract Vector<CertStatus> requestDistPoint
              (Vector<X509Certificate> certs, URL distpoint);

  protected void refresh() {

    for (X509Certificate oldCert: certStore.keySet()) {

      try {
        oldCert.checkValidity();
      } catch (CertificateException ex) {
        certStore.remove(oldCert);
      }

    }
  }

  protected void insert(X509Certificate cert, CertStatus profile) {
    for (X509Certificate oldCert: certStore.keySet()) {
      CertStatus oldProfile = certStore.get(oldCert);

      if (cert.getSubjectX500Principal().equals(cert.getIssuerX500Principal())) 
        continue;

      if (oldCert.getSubjectX500Principal()
    		     .equals(cert.getIssuerX500Principal())) {
        try {
          cert.verify(oldCert.getPublicKey());
        } catch (Exception ex) {
          continue;
        }
        profile.setValidDepth(oldProfile.getValidDepth() + 1);
        profile.setParent(oldProfile);
        oldProfile.addChild(profile);
        continue;
      }

      if (oldCert.getIssuerX500Principal()
    		  .equals(cert.getSubjectX500Principal())) {

        try {
          oldCert.verify(cert.getPublicKey());
        } catch (Exception ex) {
          continue;
        }
        oldProfile.setValidDepth(profile.getValidDepth() + 1);
        profile.addChild(oldProfile);
        oldProfile.setParent(profile);
        continue;
      }
    }
    certStore.put(cert, profile);
  }

  public CertStatus[] checkStatus(X509Certificate[] certs) {
	CryptoHelper.installBC();
	  
    Vector<CertStatus> status = new Vector<CertStatus>();
    Vector<X509Certificate> newCerts = new Vector<X509Certificate>();

    refresh();

    for (X509Certificate cert: certs) {
      

      if (newCerts.contains(cert))
        continue;

      

      try {
        cert.checkValidity();
      } catch (CertificateException ex) {
        status.add(new CertStatus(cert.getSerialNumber(),
                                  CertStatus.DEPTH_CHECKED_BAD));
        continue;
      }

      newCerts.add(cert);     

      if (!certStore.containsKey(cert)) {
        CertStatus profile = new CertStatus(cert);
        profile.addAllDistPoint(retrieveDistPoint(cert));
        insert(cert, profile);
      }
    }


    Vector<X509Certificate> uncheckedCerts = new Vector<X509Certificate>();
    uncheckedCerts.addAll(newCerts);


    for (URL dist: trustedDists) {
      Vector<CertStatus> response = requestDistPoint(uncheckedCerts, dist);
      //System.err.println(uncheckedCerts.size());
      
      for (CertStatus resp: response) {

        for (X509Certificate c: uncheckedCerts) {

          if (resp.getSerial().equals(c.getSerialNumber())) {
            
            uncheckedCerts.remove(c);
            CertStatus profile = certStore.get(c);
            profile.setValidDepth(0);
          }
        }
      }
    }



    X509Certificate firstCert = null;
    CertStatus firstProfile = null;
    while (!uncheckedCerts.isEmpty()) {

      if (firstCert == null) {
        firstCert = uncheckedCerts.remove(0);
        firstProfile = certStore.get(firstCert);
      }
      if (!firstProfile.getDistPoint().isEmpty()) {
        URL distPoint = firstProfile.getDistPoint().remove(0);
        Vector<X509Certificate> checkCerts = new Vector<X509Certificate>();
        checkCerts.add(firstCert);

        for (X509Certificate c: uncheckedCerts) {
          //System.err.println("remove " + c.getSerialNumber().toString(16));
          if (certStore.get(c).getDistPoint().remove(distPoint))
            checkCerts.add(c);
        }

        Vector<CertStatus> response = requestDistPoint(checkCerts, distPoint);

        for (CertStatus resp: response) {

          for (X509Certificate c: uncheckedCerts) {
            if (resp.getSerial().equals(c.getSerialNumber())) {
              uncheckedCerts.remove(c);
              CertStatus p = certStore.get(c);

              if (resp.getValidDepth() == CertStatus.DEPTH_CHECKED_BAD)
                p.setValidDepth(CertStatus.DEPTH_CHECKED_BAD);
              else if (p.getParent() != null)
                p.setValidDepth(p.getParent().getValidDepth());
              else
                p.setValidDepth(CertStatus.DEPTH_CHECKED_GOOD);
            }
          }

          if (resp.getSerial().equals(firstCert.getSerialNumber())) {
            CertStatus p = certStore.get(firstCert);

            if (p.getParent() != null)
              p.setValidDepth(p.getParent().getValidDepth());
            else
              p.setValidDepth(CertStatus.DEPTH_CHECKED_GOOD);

            firstCert = null;
            firstProfile = null;
          }
        }
      }
      else {
        firstCert = null;
        firstProfile = null;
      }
    }

    for (X509Certificate cert: newCerts) {
      CertStatus profile = certStore.get(cert);
      status.add(new CertStatus(cert.getSerialNumber(),
                                profile.getValidDepth()));
    }

    return status.toArray(new CertStatus[status.size()]);
  }

}
