package org.netserv.ipauth.impl;

import org.netserv.ipauth.util.CryptoHelper;
import org.netserv.ipauth.core.CertStatus;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.security.cert.X509CRL;
import java.security.cert.X509Certificate;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.DERSequence;
import org.bouncycastle.asn1.x509.DistributionPoint;
import org.bouncycastle.asn1.x509.DistributionPointName;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;
import org.bouncycastle.asn1.x509.X509Extensions;
import org.bouncycastle.jce.provider.X509CRLParser;
import org.netserv.ipauth.core.CertStatus;
import org.netserv.ipauth.core.IpAuthVerifierImpl;

/**
 * @author Chiache Tsai 
 * 
 * The class CrlVerifier is an implementation of Certificate Verifier.  
 */

public class CrlVerifier extends IpAuthVerifierImpl {
  @Override
  protected Vector<URL> retrieveDistPoint(X509Certificate cert) {
    String cdpOid = X509Extensions.CRLDistributionPoints.getId();
    byte[] cdpBytes = cert.getExtensionValue(cdpOid);

    Vector<URL> distpoints = new Vector<URL>();
    if (cdpBytes == null)
      return distpoints;
    
    Enumeration cdps = null;
    try {
      ASN1InputStream cdpIn = new ASN1InputStream
          (new ByteArrayInputStream(cdpBytes));
      cdpBytes
        = DEROctetString.getInstance(cdpIn.readObject()).getOctets();
      cdpIn = new ASN1InputStream
          (new ByteArrayInputStream(cdpBytes));
      cdps = DERSequence.getInstance(cdpIn.readObject()).getObjects();
    } catch (IOException ex) {
      return distpoints;
    }


    while (cdps.hasMoreElements()) {

      DistributionPoint cdp
          = DistributionPoint.getInstance(cdps.nextElement());
      if (cdp.getDistributionPoint() != null &&
        cdp.getDistributionPoint().getType()
        == DistributionPointName.FULL_NAME)
        try {
          GeneralName[] names = GeneralNames.getInstance
                    (cdp.getDistributionPoint().getName())
                    .getNames();
          for(GeneralName name: names) {
            distpoints.add
            (new URL(new String(CryptoHelper.getGeneralName(name))));
          }
        } catch (IOException ex) {
          continue;
        }
    }
    return distpoints;
  }

  Hashtable<X509CRL, CertStatus> crlStore
      = new Hashtable<X509CRL, CertStatus>();

  /**
   *
   */
  protected void refreshCrls() {
    Date current = new Date();
    for (X509CRL crl: crlStore.keySet()) {
      if (crl.getNextUpdate().before(current)) {
        crlStore.remove(crl);
        continue;
      }
      CertStatus cp = crlStore.get(crl);
      if (cp.getParent() == null) {
        for (X509Certificate cert: certStore.keySet()) {
          if (cert.getSubjectX500Principal()
              .equals(crl.getIssuerX500Principal())) {
            try {
              crl.verify(cert.getPublicKey());
            } catch (Exception ex) {
              continue;
            }
            CertStatus profile = certStore.get(cert);
            cp.setParent(profile);
            cp.setValidDepth(profile.getValidDepth());
            profile.addChild(cp);
          }
        }
      }
    }
  }

  /**
   *
   * @param crl
   * @param profile
   */
  protected void insertCrl(X509CRL crl, CertStatus profile) {
    for (X509Certificate oldCert: certStore.keySet()) {
      CertStatus oldProfile = certStore.get(oldCert);

      if (oldCert.getSubjectX500Principal().equals(crl.getIssuerX500Principal())) {
        try {
          crl.verify(oldCert.getPublicKey());
        } catch (Exception ex) {
          continue;
        }
        profile.setValidDepth(oldProfile.getValidDepth());
        profile.setParent(oldProfile);
        oldProfile.addChild(profile);
        continue;
      }
    }
    crlStore.put(crl, profile);
  }

  /**
   *
   * @param certs
   * @param distPoint
   * @return
   */
  @Override
  protected Vector<CertStatus> requestDistPoint
                          (Vector<X509Certificate> certs, URL distPoint) {
    refreshCrls();

    //System.err.println("Querying " + distPoint);

    Vector<X509Certificate> certsCopy = (Vector<X509Certificate>)certs.clone();
    Vector<CertStatus> status = new Vector<CertStatus>();
    for (int i = 0 ; i < certsCopy.size() ; i++) {
      X509Certificate cert = certsCopy.get(i);
      for (X509CRL crl: crlStore.keySet()) {
        CertStatus cp = crlStore.get(crl);
        if (cp.getParent() == null)
          continue;

        if (cert.getIssuerX500Principal().equals(crl.getIssuerX500Principal())) {
          X509Certificate caCert = crlStore.get(crl).getParent().getCertificate();
          try {
            cert.verify(caCert.getPublicKey());
          } catch (Exception ex) {
            continue;
          }
          
          if (crl.isRevoked(cert))
            status.add(new CertStatus(cert.getSerialNumber(), CertStatus.DEPTH_CHECKED_BAD));
          else
            status.add(new CertStatus(cert.getSerialNumber(), CertStatus.DEPTH_CHECKED_GOOD));
          certsCopy.remove(i);
          i--;
        }
      }
    }

    for (X509CRL crl: crlStore.keySet()) {
      CertStatus cp = crlStore.get(crl);
      if (cp.getDistPoint().contains(distPoint))
        return status;
    }

    try {
      HttpURLConnection urlConn = (HttpURLConnection)distPoint.openConnection();
      urlConn.setDoInput(true);
      urlConn.setDoOutput(false);
      urlConn.setRequestMethod("GET");
      urlConn.connect();

      X509CRLParser crlParser = new X509CRLParser();
      crlParser.engineInit(urlConn.getInputStream());
      Collection<X509CRL> newCrls = (Collection<X509CRL>)crlParser.engineReadAll();

      Date current = new Date();
      for (X509CRL crl: newCrls) {
        if (crl.getNextUpdate().before(current))
          continue;

        CertStatus profile = new CertStatus();
        insertCrl(crl, profile);
        if (profile.getParent() == null)
          continue;

        

        for (int i = 0 ; i < certsCopy.size() ; i++) {
          X509Certificate cert = certsCopy.get(i);
          if (cert.getIssuerX500Principal().equals(crl.getIssuerX500Principal())) {
            try {
              cert.verify(profile.getParent().getCertificate().getPublicKey());
            } catch (Exception ex) {
              continue;
            }

            

            if (crl.isRevoked(cert))
              status.add(new CertStatus(cert.getSerialNumber(), CertStatus.DEPTH_CHECKED_BAD));
            else
              status.add(new CertStatus(cert.getSerialNumber(), CertStatus.DEPTH_CHECKED_GOOD));
            certsCopy.remove(i);
            i--;
          }
        }
      }
    } catch (Exception ex) {
      ex.printStackTrace();
    }

    return status;
  }



}
