package org.netserv.ipauth.impl;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.math.BigInteger;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Properties;
import org.bouncycastle.asn1.nist.NISTObjectIdentifiers;
import org.netserv.ipauth.core.IpAuthProverImpl;
import org.netserv.ipauth.util.CryptoHelper;

/**
 *
 * @author allents
 */

public class DhcpProver extends IpAuthProverImpl {

  static public final String DEFAULT_DHCPCMD    = "udhcpc";
  static public final String DEFAULT_DHCPCMDXID = "-x";
  static public final String DEFAULT_DHCPCMDINT = "-i";
  static public final String DEFAULT_LEASEFILE  = "/var/lib/misc/udhcpd-fix.leases";
  static public final String DEFAULT_PIDFILE    = "/var/run/udhcpd-fix.pid";

  private LeaseReader reader;
  private String DHCPCmd, DHCPCmdXid, DHCPCmdInt;

  public DhcpProver(String configFile) throws IOException {
    CryptoHelper.installBC();
    
    loadUdhcpConfig(configFile);
  }

  public void loadUdhcpConfig(String configFile)
          throws IOException {
    if (configFile == null) {

      /* Client-side config */
      DHCPCmd     = DEFAULT_DHCPCMD;
      DHCPCmdXid  = DEFAULT_DHCPCMDXID;
      DHCPCmdInt  = DEFAULT_DHCPCMDINT;

      /* Server-side config */
      reader = new UdhcpLeaseReader(DEFAULT_LEASEFILE, DEFAULT_PIDFILE);
    }
    else {
     
      Properties prop = new Properties();
      prop.load(new FileInputStream(configFile));

      /* Client-side config */
      DHCPCmd     = prop.getProperty("dhcp-cmd", DEFAULT_DHCPCMD);
      DHCPCmdXid  = prop.getProperty("dhcp-cmd-xid", DEFAULT_DHCPCMDXID);
      DHCPCmdInt  = prop.getProperty("dhcp-cmd-int", DEFAULT_DHCPCMDINT);

      /* Server-side config */
      String leaseFile = prop.getProperty("lease-file", DEFAULT_LEASEFILE);
      String pidFile   = prop.getProperty("pid-file",   DEFAULT_PIDFILE);

      if (leaseFile != null && pidFile != null)
        reader = new UdhcpLeaseReader(leaseFile, pidFile);
    }
  }

  public byte[] obtain(PublicKey key, String device) {
    if (DHCPCmd == null || DHCPCmdXid == null || DHCPCmdInt == null)
      return null;

    int digestValue;
    try {
      byte[] digest = digestPublicKey(key);
      digestValue = new BigInteger(digest).intValue();
    } catch (Exception ex) { return null; }

    System.err.println("[PROVER] Generate XID = " + digestValue);

    String cmd = DHCPCmd + " " + DHCPCmdXid + " " + digestValue + " " +
                 DHCPCmdInt + " " + device;
    try {
      System.err.println("[PROVER] Run " + cmd);

      Process udhcpc = Runtime.getRuntime().exec(cmd);
      BufferedReader input = new BufferedReader
              (new InputStreamReader(udhcpc.getInputStream()));
      String line;
      while ((line = input.readLine()) != null) {
        System.err.println("[UDHCP] " + line);
      }
    } catch (Exception ex) { return null; }

    try {
      Enumeration<InetAddress> e =
              NetworkInterface.getByName(device).getInetAddresses();
      while (e.hasMoreElements()) {
        InetAddress a = e.nextElement();
        if (a instanceof Inet4Address)
          return a.getAddress();
      }
      return null;
    } catch (Exception ex) { return null; }
  }

  public boolean prove(byte[] identifier, PublicKey key) {
    
    byte[] xid = reader.getXid(identifier);
    byte[] digest;
    try {
      digest = digestPublicKey(key);
    } catch (Exception ex) {
      return false;
    }

    return Arrays.equals(xid, digest);
  }

  protected byte[] digestPublicKey(PublicKey key)
          throws NoSuchAlgorithmException, NoSuchProviderException {

    MessageDigest dig = MessageDigest.getInstance
                        (NISTObjectIdentifiers.id_sha256.getId(), "BC");
    dig.update(key.getEncoded());
    byte[] digest = dig.digest();
    byte[] partOfDigest = new byte[4];
    System.arraycopy(digest, 0, partOfDigest, 0, 4);

    return partOfDigest;
  }
}
