package org.netserv.ipauth.impl;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.SignatureException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarInputStream;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;
import sun.misc.BASE64Encoder;
import sun.security.util.ManifestDigester;
import sun.security.util.SignatureFile;

import org.netserv.ipauth.core.IpAuthSignerImpl;
import org.netserv.ipauth.util.CryptoHelper;
import org.netserv.ipauth.util.FileHelper;

/**
 * 
 * @author Chiache Tsai
 */

public class JarSigner extends IpAuthSignerImpl {
  public static int buffSize = 4096;
  public static String alias = "ipauth";
  private String algorithm;
  private String provider;
  private PrivateKey privateKey;
  private X509Certificate[] certChain;

  public JarSigner(PrivateKey key, String alg, String provider) {
    this.algorithm = alg;
    this.provider = provider;
    this.privateKey = key;
    this.certChain = new X509Certificate[0];
  }

  private static Manifest getManifestFile(JarFile jarFile) 
          throws IOException {
    if (jarFile.getJarEntry("META-INF/MANIFEST.MF") == null) 
      return null;
    
    Enumeration entries = jarFile.entries();
      
    JarEntry manifestEntry = null;
    while(entries.hasMoreElements()) {
      JarEntry entry = (JarEntry)entries.nextElement();

      if("META-INF/MANIFEST.MF".equalsIgnoreCase(entry.getName())) {
        manifestEntry = entry;
        break;
      }
    }

    Manifest manifest = new Manifest();
    if(manifestEntry != null)
      manifest.read(jarFile.getInputStream(manifestEntry));
      
    return manifest;
  }

  private static Map createEntries(Manifest manifest, JarFile jarFile)
          throws IOException {
    Map entries = manifest.getEntries();

    if(entries.size() > 0) {
      Iterator elements = entries.keySet().iterator();

      while(elements.hasNext()) {
        String element = (String)elements.next();
        if(jarFile.getEntry(element) == null)
          elements.remove();
      }
    }
    
    return entries;
  }

  private static BASE64Encoder b64Encoder = new BASE64Encoder();

  private static Map updateManifestEntries(Manifest manifest, 
                                           JarFile jarFile,
                                           MessageDigest messageDigest, 
                                           Map entries) 
          throws IOException {
    Enumeration jarElements = jarFile.entries();
    
    while(jarElements.hasMoreElements()) {
      JarEntry jarEntry = (JarEntry)jarElements.nextElement();
      if(jarEntry.getName().startsWith("META-INF"))
        continue;
      
      InputStream jarInput = jarFile.getInputStream(jarEntry);
      byte[] buffer = new byte[buffSize];
      int read = 0;
      while((read = jarInput.read(buffer)) > 0)
        messageDigest.update(buffer, 0, read);
      jarInput.close(); 
      String digest = b64Encoder.encode(messageDigest.digest());
      String digestAlg = messageDigest.getAlgorithm().concat("-Digest");
      
      if(manifest.getAttributes(jarEntry.getName()) != null) {
        Attributes attributes = manifest.getAttributes(jarEntry.getName());
        attributes.putValue(digestAlg, digest);

      } else if(!jarEntry.isDirectory()) {
        Attributes attributes = new Attributes();
        attributes.putValue(digestAlg, digest);
        entries.put(jarEntry.getName(), attributes);
      }
    }
    
    return entries;
  }

  private SignatureFile createSignatureFile(Manifest manifest,
		                                    MessageDigest messageDigest) 
          throws IOException {
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    manifest.write(baos);
    baos.flush();
    baos.close();
    byte[] byteArray = baos.toByteArray();

    ManifestDigester manifestDigester = new ManifestDigester(byteArray);

    byte[] digest = manifestDigester.manifestDigest(messageDigest);
    return new SignatureFile(new MessageDigest[] { messageDigest }, 
                             manifest, manifestDigester, alias, true);
  }

  private static void writeJarEntry(JarEntry entry, JarFile jarFile, 
                                    JarOutputStream jos)
          throws IOException {
    jos.putNextEntry(entry);
    byte[] buffer = new byte[2048];
    int read = 0;
    InputStream is = jarFile.getInputStream(entry);
    while((read = is.read(buffer)) > 0)
        jos.write(buffer, 0, read);
    jos.closeEntry();
  }
 
  @Override
  public void addCaPub(X509Certificate caPub) {
    X509Certificate[] newCertChain = new X509Certificate[certChain.length + 1];
    for (int i = 0 ; i < certChain.length ; i++) {
    	newCertChain[i] = certChain[i];
    }
    newCertChain[certChain.length] = caPub;
    certChain = newCertChain;
  }

  @Override
  public void sign(InputStream input, OutputStream output) throws IOException { 
    File tmpFile = File.createTempFile("signing-", ".jar"); 
	FileOutputStream tmpOutput = new FileOutputStream(tmpFile);
	byte buffer[] = new byte[buffSize];
    int read = 0;
    while((read = input.read(buffer)) > 0)
      tmpOutput.write(buffer, 0, read);
    tmpOutput.close();
    input.close();
    
    JarFile jarFile = new JarFile(tmpFile);
    
    Manifest manifest = getManifestFile(jarFile);
    Map entries = createEntries(manifest, jarFile);
    
    MessageDigest messageDigest;
    try {
      messageDigest = MessageDigest.getInstance(algorithm);
    } catch (NoSuchAlgorithmException e) {
      e.printStackTrace();
      return;
    }
    
    updateManifestEntries(manifest, jarFile, messageDigest, entries);
    
    SignatureFile signatureFile;
    SignatureFile.Block block;
    
    signatureFile = createSignatureFile(manifest, messageDigest);
    try {
		block = signatureFile.generateBlock(privateKey, certChain, true);
	} catch (InvalidKeyException e) {
		e.printStackTrace();
		return;
	} catch (NoSuchAlgorithmException e) {
		e.printStackTrace();
		return;
	} catch (SignatureException e) {
		e.printStackTrace();
		return;
	} catch (CertificateException e) {
		e.printStackTrace();
		return;
	}
    
    String manifestFileName = "META-INF/MANIFEST.MF";
    JarOutputStream jos = new JarOutputStream(output);
    JarEntry manifestFile = new JarEntry(manifestFileName);
    jos.putNextEntry(manifestFile);
    manifest.write(jos);
    jos.closeEntry();

    String signatureFileName = signatureFile.getMetaName();
    JarEntry signatureFileEntry = new JarEntry(signatureFileName);
    jos.putNextEntry(signatureFileEntry);
    signatureFile.write(jos);
    jos.closeEntry();

    String signatureBlockName = block.getMetaName();
    JarEntry signatureBlockEntry = new JarEntry(signatureBlockName);
    jos.putNextEntry(signatureBlockEntry);
    block.write(jos);
    jos.closeEntry();

    Enumeration metaEntries = jarFile.entries();
    while(metaEntries.hasMoreElements()) {
      JarEntry metaEntry = (JarEntry)metaEntries.nextElement();
      String metaName = metaEntry.getName();
      if (metaName.startsWith("META-INF") &&
          !metaName.equalsIgnoreCase(manifestFileName)  &&
          !metaName.equalsIgnoreCase(signatureFileName) &&
          !metaName.equalsIgnoreCase(signatureBlockName))
          writeJarEntry(metaEntry, jarFile, jos);
    }

    Enumeration allEntries = jarFile.entries();
    while(allEntries.hasMoreElements()) {
      JarEntry entry = (JarEntry)allEntries.nextElement();
      if(!entry.getName().startsWith("META-INF"))
        writeJarEntry(entry, jarFile, jos);
    }

    jos.flush();
    jos.finish();
    jarFile.close();
  }
}
