package org.netserv.ipauth.service;

import com.novosec.pkix.asn1.cmp.PKIMessage;
import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpServer;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.NoSuchProviderException;
import java.security.Signature;
import java.security.cert.CRLException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509CRL;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.Properties;
import java.util.Random;

import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.nist.NISTObjectIdentifiers;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
import org.bouncycastle.asn1.x509.X509CertificateStructure;
import org.bouncycastle.asn1.x509.X509Name;
import org.netserv.ipauth.util.CmpHelper;
import org.netserv.ipauth.util.CryptoHelper;

/**
 * Create a RA/CA server for the PKI system.
 * 
 * @author Chia-che Tsai
 */

public class RaServer {
  private static final String DEFAULT_CONFIGFILE = "ra.properties";

  private static final String DEFAULT_RA_BIND = "0.0.0.0";
  private static final String DEFAULT_RA_PORT = "4000";
  private static final String DEFAULT_DISTPOINT  = "/ra";

  private HttpServer server;

  private InetSocketAddress myAddress  = null;
  private String distPoint             = null;
  private URL    caCmpUrl    = null;
  private String caSecret    = null;
  private String caCertDist  = null;
  private X509Certificate caCert = null;
  private X509CertificateStructure caCertStruct = null;
  private String caCrlDist   = null;
  private X509CRL caCrl      = null;

  /**
   *
   * @param configFile
   * @throws IOException
   */
  public RaServer(String configFile) throws IOException {
    CryptoHelper.installBC();

    server = HttpServer.create();

    loadConfig(configFile);
  }

  /**
   *
   * @throws MalformedURLException
   * @throws IOException
   * @throws CertificateException
   * @throws NoSuchProviderException
   */
  public void loadCaCert()
      throws MalformedURLException, IOException,
           CertificateException, NoSuchProviderException {

    CertificateFactory certFact = CertificateFactory.getInstance("X.509", "BC");

    InputStream is = null;

    if (caCertDist.startsWith("http://")) {

      URL caCertDistUrl = new URL(caCertDist);
      HttpURLConnection urlConn = (HttpURLConnection)caCertDistUrl.openConnection();
      urlConn.setDoInput(true);
      urlConn.setDoOutput(false);
      urlConn.setRequestMethod("GET");
      urlConn.connect();

      is = urlConn.getInputStream();
    }
    else
      is = new FileInputStream(caCertDist);

    caCert = (X509Certificate)certFact.generateCertificate(is);
    caCertStruct = CryptoHelper.X509CertificateToX509CertificateStructure(caCert);
    System.err.println("[RA] load CA Certificate");
  }

  /**
   *
   * @throws MalformedURLException
   * @throws IOException
   * @throws CertificateException
   * @throws NoSuchProviderException
   * @throws CRLException
   */
  public void loadCaCrl()
      throws MalformedURLException, IOException,
           CertificateException, NoSuchProviderException,
           CRLException {

    CertificateFactory certFact = CertificateFactory.getInstance("X.509", "BC");


    InputStream is = null;

    if (caCrlDist.startsWith("http://")) {

      URL caCrlDistUrl = new URL(caCrlDist);
      HttpURLConnection urlConn = (HttpURLConnection)caCrlDistUrl.openConnection();
      urlConn.setDoInput(true);
      urlConn.setDoOutput(false);
      urlConn.setRequestMethod("GET");
      urlConn.connect();

      is = urlConn.getInputStream();
    }
    else
      is = new FileInputStream(caCrlDist);

    caCrl = (X509CRL)certFact.generateCRL(is);

    System.err.println("[RA] load CA CRL");
  }

  /**
   *
   * @param configFile
   * @throws IOException
   */
  public void loadConfig(String configFile) throws IOException {
    if (configFile == null)
      configFile = DEFAULT_CONFIGFILE;

    Properties raProp = new Properties();
    raProp.load(new FileInputStream(configFile));


    String bind = raProp.getProperty("ra-bind", DEFAULT_RA_BIND);
    String port = raProp.getProperty("ra-port", DEFAULT_RA_PORT);
    try {
      myAddress = new InetSocketAddress(InetAddress.getByName(bind), Integer.parseInt(port));
    } catch (Exception ex) {
      myAddress = null;
    }
    if (myAddress == null) {
      System.err.println("[RA] cannot recognize binding address");
      return;
    }

    distPoint = raProp.getProperty("ra-distpoint", DEFAULT_DISTPOINT);


    String cmpUrl = raProp.getProperty("ca-cmp-url");
    String secret = raProp.getProperty("ca-cmp-secret");
    if (cmpUrl == null || secret == null){
      System.err.println("[RA] cannot load CA cmp connection");
      return;
    }
    caCmpUrl = new URL(cmpUrl);
    caSecret = secret;

    caCertDist = raProp.getProperty("ca-cert");
    try {
      if (caCertDist == null)
        throw new NullPointerException();
      loadCaCert();
    } catch (Exception ex) {
      ex.printStackTrace();
      System.err.println("[RA] cannot recognize CA certificate");
      return;
    }

    caCrlDist  = raProp.getProperty("ca-crl");
    try {
      if (caCrlDist == null)
        throw new NullPointerException();
      loadCaCrl();
    } catch (Exception ex) {
      ex.printStackTrace();
      System.err.println("[RA] cannot recognize CA revocation list");
      return;
    }
  }

  /**
   *
   */
  public void start() {
    try {
      server.createContext(distPoint);
      server.bind(myAddress, 0);
      System.err.println("[RA] bind the distribute point to http://" +
          myAddress.getHostName() + ":" + myAddress.getPort() + distPoint);
      server.start();
    } catch (IOException ex) {
      ex.printStackTrace();
      server.stop(0);
    }
  }

  public void handle(HttpExchange he) throws IOException {

    if (he.getRequestMethod().equals("POST")) {
      try {
        Date current = new Date();
        if (caCert.getNotAfter().before(current))
          loadCaCert();
        if (caCrl.getNextUpdate().before(current))
          loadCaCrl();
      } catch (Exception ex) {
        he.sendResponseHeaders(405, 0);
        OutputStream responseBody = he.getResponseBody();
        responseBody.write("Server not validated".getBytes());
        responseBody.flush();
        return;
      }

      ASN1InputStream asn1is = new ASN1InputStream(he.getRequestBody());

      PKIMessage pkireq;
      try {
        pkireq = PKIMessage.getInstance(asn1is.readObject());
        if (pkireq == null)
          throw new CertificateException("no PKI message");

        AlgorithmIdentifier algId = pkireq.getHeader().getProtectionAlg();

        byte[] protectedBytes = pkireq.getProtectedBytes();
        byte[] protection = pkireq.getProtection().getBytes();

        X509CertificateStructure masterCertStruct = pkireq.getExtraCert(0);
        if (protection == null || masterCertStruct == null)
          throw new CertificateException("no certificate protection");

        X509Name subject = masterCertStruct.getSubject();
        System.err.println("[RA] Receive a message from " + subject);

        X509Certificate masterCert = CryptoHelper.X509CertificateStructureToX509Certificate
                       (masterCertStruct);
        if (!masterCert.getIssuerX500Principal().equals(caCert.getSubjectX500Principal()))
          throw new CertificateException("unmatched master cetificate");

        try {
          masterCert.checkValidity();
          masterCert.verify(caCert.getPublicKey());
        } catch (Exception ex) {
          throw new CertificateException("master certificate unverified");
        }

        if (caCrl.isRevoked(masterCert))
          throw new CertificateException("master certificate revoked");
        
        try {
          Signature sig = Signature.getInstance(algId.getObjectId().getId());

          sig.initVerify(masterCert.getPublicKey());
          sig.update(protectedBytes);
          if (!sig.verify(protection))
            throw new Exception();
        } catch (Exception ex) {
          throw new CertificateException("cmp message not validated");
        }

        try {
          pkireq = CmpHelper.protectPKIMessageWithPBE
               (pkireq, "ra", caSecret,
                NISTObjectIdentifiers.id_sha256.getId(),
                PKCSObjectIdentifiers.id_hmacWithSHA256.getId(),
                new Random().nextInt(100));
        } catch (Exception ex) {
          if (ex instanceof CertificateException)
            throw (CertificateException)ex;
          else
            throw new CertificateException("cannot access service");
        }
      } catch (CertificateException ex) {
        System.err.println("[RA] A connection are being denied");
        he.sendResponseHeaders(405, 0);
        OutputStream responseBody = he.getResponseBody();
        responseBody.write("Invalid request: ".getBytes());
        responseBody.write(ex.getMessage().getBytes());
        responseBody.close();
        return;
      }

      System.err.println("[RA] Send the request to CA");
      HttpURLConnection urlConn = (HttpURLConnection)caCmpUrl.openConnection();
      urlConn.setDoInput(true);
      urlConn.setDoOutput(true);
      urlConn.setRequestMethod("POST");

      OutputStream httpOut = urlConn.getOutputStream();
      httpOut.write(pkireq.getDERObject().getEncoded());
      httpOut.close();

      InputStream httpIn = urlConn.getInputStream();
      PKIMessage pkimsgRep = PKIMessage.getInstance(
                   new ASN1InputStream(httpIn).readObject());
      pkimsgRep.addExtraCert(caCertStruct);
      System.err.println("[RA] recevie response from CA");
      System.err.println("[RA] Send the response to client");
      he.sendResponseHeaders(200, 0);
      OutputStream responseBody = he.getResponseBody();
      responseBody.write(pkimsgRep.getDERObject().getEncoded());
      responseBody.close();

    }
    else {
      System.err.println("[RA] A connection are being denied");
      he.sendResponseHeaders(405, 0);
      OutputStream responseBody = he.getResponseBody();
      responseBody.write("You can only use 'POST' method".getBytes());
      responseBody.close();
    }
  }
}
