package org.netserv.ipauth.sign;

import java.io.ByteArrayInputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.security.Signature;
import java.security.cert.CertStore;
import java.security.cert.CertStoreException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.cms.SignedData;
import org.bouncycastle.asn1.x509.X509CertificateStructure;
import org.bouncycastle.cms.CMSException;
import org.bouncycastle.cms.CMSSignedData;
import org.bouncycastle.cms.CMSSignedDataParser;
import org.bouncycastle.cms.SignerInformation;

import org.netserv.ipauth.core.IpAuthVerifier;
import org.netserv.ipauth.core.IpAuthVerifierImpl;
import org.netserv.ipauth.util.CryptoHelper;

/**
*
* @author Chia-che Tsai
*/

public class SignedInputStream extends InputStream {
  private ASN1InputStream      asn1is;
  private ByteArrayInputStream bais;
  private IpAuthVerifier       verifier;

  private class SignedMessage extends Object {
    private byte[] data;
    private X509Certificate[] certChain;
    
    @Override
    public String toString() {
      String str = "Message:\n" + new String(data) + "\n\n";
      for (int i = 0 ; i < certChain.length ; i++) {
        str += "Certificate[" + i + "]:\n" + certChain[i].toString();
      }
      return str;
    }
    
    public void setData(byte[] data) {
      this.data = data;
    }
    
    public void setCertChain(X509Certificate[] certChain) {
      this.certChain = certChain;
    } 
  }
  
  private SignedMessage lastMessage;
  
  public Object getLastMessage() {
    return lastMessage;
  }
  
  /**
   *
   * @param is
   * @param sourceAddr
   * @param verifier
   * @throws IOException
   */
  public SignedInputStream(InputStream is, IpAuthVerifier verifier)
      throws IOException {
    this.asn1is = new ASN1InputStream(is);
    this.bais = new ByteArrayInputStream(new byte[0]);
    this.verifier = verifier;
  }

  @Override
  public int read() throws IOException {
    if (bais.available() > 0) return bais.read();

    while(true) {
      try {
        getNextBlock();
      } catch (EOFException e) {
        e.printStackTrace();
        break;
      } catch (CMSException e) {
        e.printStackTrace();
        return 0;
      } catch (CertificateException e) {
        e.printStackTrace();
        return -1;
      }
    }
    return bais.read();
  }

  @Override
  public int read(byte[] b) throws IOException {
    return read(b, 0, b.length);
  }

  @Override
  public int read(byte[] b, int off, int len) throws IOException {
    if (len > b.length - off)
      len = b.length - off;

    if (bais.available() == 0) {
      try {
        getNextBlock();
      } catch (EOFException e) {
        e.printStackTrace();
    	return 0;
      } catch (CMSException e) {
    	e.printStackTrace();
  		return 0;
	  } catch (CertificateException e) {
	    e.printStackTrace();
	    return -1;
	  }
    }

    return bais.read(b, off, len);
  }

  /**
   *
   * @throws IOException
 * @throws CMSException 
   */
  protected void getNextBlock() 
            throws IOException, CMSException, CertificateException {

	CMSSignedDataParser sp = new CMSSignedDataParser(asn1is);

    InputStream dataIn = sp.getSignedContent().getContentStream();
    byte[] data = new byte[dataIn.available()];
    dataIn.read(data);
    
    CertStore store;
	try {
		store = sp.getCertificatesAndCRLs("Collection", "BC");
	} catch (NoSuchAlgorithmException e) {
		throw new CMSException(e.getMessage());
	} catch (NoSuchProviderException e) {
		throw new CMSException(e.getMessage());
	} 
	
    Iterator signers = sp.getSignerInfos().getSigners().iterator();
    Vector<X509Certificate> certs = new Vector<X509Certificate>();
    while (signers.hasNext()) {
      SignerInformation signer = (SignerInformation)signers.next();
      Collection certCollection;
	  try {
	    certCollection = store.getCertificates(signer.getSID());
	  } catch (CertStoreException e) {
		throw new CMSException(e.getMessage());
	  }

      Iterator certIt = certCollection.iterator();
      X509Certificate cert = (X509Certificate)certIt.next();
      try {
		signer.verify(cert, "BC");
	  } catch (NoSuchAlgorithmException e) {
		throw new CMSException(e.getMessage());
	  } catch (NoSuchProviderException e) {
		throw new CMSException(e.getMessage());
	  }
      certs.add(cert);
    }
    
    X509Certificate[] certArray = new X509Certificate[certs.size()];
    
    verifier.verify(certs.toArray(certArray));
    
    System.err.println("[SOCKET] Receive a signed message");

    lastMessage.setData(data);
    lastMessage.setCertChain(certArray);
    
    if (data.length == 0) {
      System.err.println("[SOCKET] No Payload - drop it!");
      getNextBlock();
      return;
    }
    
    bais = new ByteArrayInputStream(data);
  }
}
