package org.netserv.ipauth.ui;


import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.URL;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.Properties;
import java.util.Random;
import java.util.Vector;

import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.DERInteger;
import org.bouncycastle.asn1.DERNull;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
import org.bouncycastle.asn1.x509.X509CertificateStructure;
import org.bouncycastle.asn1.x509.X509Name;

import com.novosec.pkix.asn1.cmp.PKIBody;
import com.novosec.pkix.asn1.cmp.PKIHeader;
import com.novosec.pkix.asn1.cmp.PKIMessage;
import com.novosec.pkix.asn1.crmf.CertReqMessages;
import com.novosec.pkix.asn1.crmf.CertReqMsg;
import com.novosec.pkix.asn1.crmf.CertRequest;
import com.novosec.pkix.asn1.crmf.CertTemplate;
import com.novosec.pkix.asn1.crmf.ProofOfPossession;

import org.netserv.ipauth.core.IpAuthProver;
import org.netserv.ipauth.core.IpAuthProverImpl;
import org.netserv.ipauth.core.IpAuthProverImpl;
import org.netserv.ipauth.impl.DhcpProver;
import org.netserv.ipauth.util.CmpHelper;
import org.netserv.ipauth.util.CmpHelper;
import org.netserv.ipauth.util.CryptoHelper;
import org.netserv.ipauth.util.CryptoHelper;
import org.netserv.ipauth.util.FileHelper;
import org.netserv.ipauth.util.FileHelper;

/**
 *
 * @author Chiache Tsai
 */
public class IpAuthClient {
  /**
   * Default configuration file path (= <code>./client properties</code>)
   */
  static public final String DEFAULT_CONFIGFILE = "client.properties";
  /**
   * Default created key size (= <i>1024 bit</i>)
   */
  static public final String DEFAULT_KEYSIZE  = "1024";
  /**
   * 
   */
  static public final String DEFAULT_KEYFILE  = "localkey.pem";
  /**
   *
   */
  static public final String DEFAULT_CERTFILE = "localcert.pem";

  private String intf;

  private int keySize        = 0;
  private String keyFileStr  = null;
  private String certFileStr = null;

  private URL proverUrl = null;
  private URL caCmpUrl  = null;

  private IpAuthProver prover;

  private KeyPair     key     = null;
  private InetAddress address = null;
  private PKIMessage  pkimsg  = null;
  private X509Certificate[] certificates = null;

  /**
   *
   * @param configFile
   * @throws IOException
   */
  public IpAuthClient(String configFile) throws IOException  {
    CryptoHelper.installBC();
    loadConfig(configFile);
  }

  /**
   *
   * @param configFile
   * @throws IOException
   */
  public void loadConfig(String configFile) throws IOException {
    if (configFile == null)
      configFile = DEFAULT_CONFIGFILE;

    Properties prop = new Properties();
    prop.load(new FileInputStream(configFile));

    intf = prop.getProperty("interface");

    keySize = Integer.parseInt(prop.getProperty("key-size",  DEFAULT_KEYSIZE));
    keyFileStr  = prop.getProperty("key-file",  DEFAULT_KEYFILE);
    certFileStr = prop.getProperty("cert-file", DEFAULT_CERTFILE);

    proverUrl  = new URL(prop.getProperty("prover-url"));

    String proverName   = prop.getProperty("prover");
    String proverConfig = prop.getProperty("prover-config");
    
    PrivateKey key = FileHelper.readKeyPair(keyFileStr, null).getPrivate();
    
    if (proverName.equalsIgnoreCase("udhcp")) {
      prover = new IpAuthProver(key, new DhcpProver(proverConfig));
    }
    else {
      return;
    }

    caCmpUrl  = new URL(prop.getProperty("ca-cmp-url"));
  }

  /**
   *
   * @throws Exception
   */
  public void generateKey() throws Exception {
    try {
      KeyPairGenerator keyGen = KeyPairGenerator.getInstance
              (PKCSObjectIdentifiers.rsaEncryption.getId(), "BC");
      keyGen.initialize(keySize);
      key = keyGen.genKeyPair();
    } catch (Exception ex) { throw new Exception(); }
  }

  /**
   *
   * @throws Exception
   */
  public void obtainIP() throws Exception {
    try {
      address = InetAddress.getByAddress(prover.obtain(key.getPublic(), intf));
    } catch (Exception ex) { throw new Exception(); }
  }

  /**
   *
   * @throws Exception
   */
  public void requestProver() throws Exception {
    SubjectPublicKeyInfo keyinfo;
    try {
      keyinfo = CryptoHelper.PublicKeyToSubjectPublicKeyInfo(key.getPublic());
    } catch (Exception ex) { throw new Exception(); }

    CertTemplate temp = new CertTemplate();
    temp.setPublicKey(keyinfo);
    temp.setSubject(new X509Name("CN=" + address.getHostAddress()));

    CertRequest reqinfo =
            new CertRequest(new DERInteger(new Random().nextInt()), temp);
    CertReqMsg reqmsg = new CertReqMsg(reqinfo);
    reqmsg.setPop(new ProofOfPossession(new DERNull(), 0));

    CertReqMessages reqmsgs = new CertReqMessages(reqmsg);

    PKIHeader header = CmpHelper.createPKIHeader(temp.getSubject(), 
                                                 new X509Name("CN=server"));
    PKIMessage message = new PKIMessage(header, new PKIBody(reqmsgs, 0));
    PKIMessage signedMessage = null;

    try {
      HttpURLConnection urlConn = (HttpURLConnection)proverUrl.openConnection();
      urlConn.setDoInput(true);
      urlConn.setDoOutput(true);
      urlConn.setRequestMethod("POST");

      OutputStream httpOut = urlConn.getOutputStream();
      httpOut.write(message.getDERObject().getEncoded());
      httpOut.flush();

      InputStream httpIn = urlConn.getInputStream();
      signedMessage =
              PKIMessage.getInstance(new ASN1InputStream(httpIn).readObject());
    } catch (Exception ex) { throw new Exception(); }

    pkimsg = signedMessage;
  }

  /**
   *
   * @throws Exception
   */
  public void requestCertificate() throws Exception {
    try {
      HttpURLConnection urlConn = (HttpURLConnection)caCmpUrl.openConnection();
      urlConn.setDoInput(true);
      urlConn.setDoOutput(true);
      urlConn.setRequestMethod("POST");

      OutputStream httpOut = urlConn.getOutputStream();
      httpOut.write(pkimsg.getDERObject().getEncoded());
      httpOut.flush();

      InputStream httpIn = urlConn.getInputStream();
      PKIMessage pkirep =
              PKIMessage.getInstance(new ASN1InputStream(httpIn).readObject());

      X509Certificate cert =
             CryptoHelper.X509CertificateStructureToX509Certificate(
             pkirep.getBody().getIp().getResponse(0).getCertifiedKeyPair().
             getCertOrEncCert().getCertificate());

      Vector<X509Certificate> certs = new Vector<X509Certificate>();
      certs.add(cert);
      for (int i = 0 ; ; i++) {
        X509CertificateStructure extraCert = pkirep.getExtraCert(i);
        if (extraCert == null)
          break;
        certs.add
            (CryptoHelper.X509CertificateStructureToX509Certificate(extraCert));
      }

      certificates = certs.toArray(new X509Certificate[certs.size()]);
    } catch (Exception ex) {
      ex.printStackTrace();
      throw new Exception(); }
  }

  /**
   *
   */
  public void storeFile() {
    try {
      FileHelper.writeCertificates(certificates, certFileStr);
      System.err.println("[Client] Certificate cached to " + certFileStr);
      FileHelper.writeKeyPair(key, this.keyFileStr);
      System.err.println("[Client] Private Key cached to " + keyFileStr);
    } catch (IOException ex) {
      System.err.println("[Client] File cache failed");
    }
  }

  /**
   *
   */
  public void request() {

    try {
      generateKey();
    } catch(Exception ex) {
      System.err.println("[Client] Key Generation failed.");
      return;
    }

    System.err.println("[Client] Key Generated");

    try {
      obtainIP();
    } catch(Exception ex) {
      System.err.println("[Client] IP Obtaining failed.");
      return;
    }

    System.err.println("[Client] Get IP: " + address.getHostAddress());

    try {
      requestProver();
    } catch(Exception ex) {
      System.err.println("[Client] Request Prover failed.");
      return;
    }

    System.err.println("[Client] Got IP ownership proved");


    try {
      requestCertificate();
    } catch(Exception ex) {
      ex.printStackTrace();
      System.err.println("[Client] Request Certificate failed.");
      return;
    }

    System.err.println("[Client] Receive Certificate from CA");

    storeFile();
  }

  /**
   * 
   * @return
   */
  public KeyPair getKey() {
    return key;
  }

  /**
   *
   * @return
   */
  public InetAddress getAddress() {
    return address;
  }

  /**
   *
   * @return
   */
  public PKIMessage getPkiMessage() {
    return pkimsg;
  }

  /**
   *
   * @return
   */
  public X509Certificate[] getCertificates() {
    return certificates;
  }
}
