package org.netserv.ipauth.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Date;
import java.util.Random;
import java.io.ByteArrayInputStream;
import java.io.UnsupportedEncodingException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.SignatureException;
import javax.crypto.Mac;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import org.bouncycastle.asn1.DERGeneralizedTime;
import org.bouncycastle.asn1.DERInteger;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.DEROutputStream;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.X509Name;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.DERBitString;
import org.bouncycastle.asn1.DERObjectIdentifier;
import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
import org.bouncycastle.asn1.x509.X509CertificateStructure;

import com.novosec.pkix.asn1.cmp.PKIHeader;
import com.novosec.pkix.asn1.cmp.PKIMessage;
import com.novosec.pkix.asn1.crmf.PBMParameter;
import com.novosec.pkix.asn1.cmp.CMPObjectIdentifiers;
import com.novosec.pkix.asn1.cmp.Challenge;
import java.security.PublicKey;
import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;

/**
 * Class <code>CmpHelper</code> provides convenient utilities for generating
 * CMP messages.
 * 
 * @author Chiache Tsai
 */
public class CmpHelper {

  /**
   *
   * @param sender
   * @param recipient
   * @param senderKID
   * @param recipientKID
   * @param senderNonce
   * @param recipientNonce
   * @param transactionId
   * @return
   */
  public static PKIHeader createPKIHeader(X509Name sender, X509Name recipient) {
	PKIHeader header = new PKIHeader(
        new DERInteger(2),
        new GeneralName(sender),
        new GeneralName(recipient));

    header.setMessageTime(new DERGeneralizedTime(new Date()));

    return header;
  }

  /**
   *
   * @param msg
   * @param keyId
   * @param raSecret
   * @param digestAlgId
   * @param macAlgId
   * @param iterationCount
   * @return
   * @throws NoSuchAlgorithmException
   * @throws NoSuchProviderException
   * @throws InvalidKeyException
   * @throws IOException
   */
  public static PKIMessage protectPKIMessageWithPBE
      (PKIMessage msg, String keyId,
       String raSecret, String digestAlgId, String macAlgId, int iterationCount)
      throws NoSuchAlgorithmException, NoSuchProviderException,
           InvalidKeyException, IOException {

    // Create the PasswordBased protection of the message
    byte[] keyIdBytes;
    try {
      keyIdBytes = keyId.getBytes("UTF-8");
    } catch (UnsupportedEncodingException e) {
      keyIdBytes = keyId.getBytes();
    }

    msg.getHeader().setSenderKID(new DEROctetString(keyIdBytes));

    AlgorithmIdentifier owfAlg = new AlgorithmIdentifier(digestAlgId);

    DERInteger iteration = new DERInteger(iterationCount);

    AlgorithmIdentifier macAlg = new AlgorithmIdentifier(macAlgId);

    byte[] saltbytes = createNonce();
    DEROctetString derSalt = new DEROctetString(saltbytes);

    String objectId = CMPObjectIdentifiers.passwordBasedMac.getId();
    PBMParameter pp = new PBMParameter(derSalt, owfAlg, iteration, macAlg);
    AlgorithmIdentifier pAlg
            = new AlgorithmIdentifier(new DERObjectIdentifier(objectId), pp);

    PKIHeader newHeader = msg.getHeader();
    newHeader.setProtectionAlg(pAlg);
    msg = new PKIMessage(newHeader, msg.getBody());

    byte[] rasecret = raSecret.getBytes();
    byte[] basekey = new byte[rasecret.length + saltbytes.length];
    for (int i = 0; i < rasecret.length; i++) {
      basekey[i] = rasecret[i];
    }

    for (int i = 0; i < saltbytes.length; i++) {
      basekey[rasecret.length+i] = saltbytes[i];
    }

    MessageDigest dig
            = MessageDigest.getInstance(owfAlg.getObjectId().getId(), "BC");

    for (int i = 0; i < iterationCount; i++) {
      basekey = dig.digest(basekey);
      dig.reset();
    }
    // Do the mac
    String macOid = macAlg.getObjectId().getId();
    byte[] protectedBytes = msg.getProtectedBytes();
    Mac mac = Mac.getInstance(macOid, "BC");
    SecretKey key = new SecretKeySpec(basekey, macOid);
    mac.init(key);
    mac.reset();
    mac.update(protectedBytes);

    // Finally store the protection bytes in the msg
    msg.setProtection(new DERBitString(mac.doFinal()));

    return msg;
  }

  /**
   *
   * @param msg
   * @param cert
   * @param key
   * @param digestAlg
   * @return
   * @throws NoSuchProviderException
   * @throws NoSuchAlgorithmException
   * @throws SecurityException
   * @throws SignatureException
   * @throws InvalidKeyException
   */
  public static PKIMessage protectPKIMessageWithCert(PKIMessage msg,
      X509CertificateStructure cert, PrivateKey key, String digestAlg)
      throws NoSuchProviderException, NoSuchAlgorithmException,
           SecurityException, SignatureException, InvalidKeyException {

    DERObjectIdentifier oid = new DERObjectIdentifier(digestAlg);
    AlgorithmIdentifier algId = new AlgorithmIdentifier(oid);
    PKIHeader newHeader = msg.getHeader();
    newHeader.setProtectionAlg(algId);
    msg = new PKIMessage(newHeader, msg.getBody());

    Signature sig = Signature.getInstance(oid.getId() , "BC");
    sig.initSign(key);
    sig.update(msg.getProtectedBytes());

    msg.setProtection(new DERBitString(sig.sign()));
	msg.addExtraCert(cert);

    return msg;
  }

  /**
   *
   * @param msg
   * @return
   * @throws IOException
   */
  public static byte[] pkiMessageToByteArray(PKIMessage msg)
      throws IOException {
    // Return response as byte array
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    DEROutputStream mout = new DEROutputStream( baos );
    mout.writeObject(msg);
    mout.close();
    return baos.toByteArray();
  }

  /**
   *
   * @return
   */
  public static byte[] createNonce() {
    // Sendernonce is a random number
    byte[] senderNonce = new byte[16];
    Random randomSource;
    randomSource = new Random();
    randomSource.nextBytes(senderNonce);
    return senderNonce;
  }

  /**
   *
   * @param message
   * @return
   */
  public static PKIMessage parsePkiMessage(byte[] message) {
    try {
      return PKIMessage.getInstance(
            new ASN1InputStream(new ByteArrayInputStream(message))
            .readObject());
    } catch (Exception e) {
        return null;
    }
  }


}
