package org.netserv.ipauth.util;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.InvalidKeyException;
import java.security.KeyException;
import java.security.KeyFactory;
import java.security.Principal;
import java.security.PublicKey;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.spec.X509EncodedKeySpec;
import java.util.Random;
import java.util.Vector;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.DERBitString;
import org.bouncycastle.asn1.DERIA5String;
import org.bouncycastle.asn1.DERObjectIdentifier;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.nist.NISTObjectIdentifiers;
import org.bouncycastle.asn1.oiw.OIWObjectIdentifiers;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
import org.bouncycastle.asn1.x509.X509CertificateStructure;
import org.bouncycastle.asn1.x509.X509Name;

/**
 * This helper provided utilities for converting cryptographic objects between
 * java.security spec and Bouncy Castle spec.
 * 
 * @author Chia-che Tsai
 */

public class CryptoHelper {

  /**
   * Installing Bouncy Castle Provider
   */
  static public void installBC() {
    java.security.Security.addProvider
    (new org.bouncycastle.jce.provider.BouncyCastleProvider());
  }

  /**
   * Convert org.bouncycastle.asn1.x509.X509CertificateStructure objects to
   * java.security.cert.X509Certificate object.
   *
   * @param certstr X509Certificate object
   * @return X509CertificateStructure object
   * @throws CertificateException if the object cannot be decoded
   */
  static public X509Certificate X509CertificateStructureToX509Certificate
      (X509CertificateStructure certstr) throws CertificateException {

    try {
      return (X509Certificate)CertificateFactory.getInstance("X.509")
              .generateCertificate(new ByteArrayInputStream(certstr.getEncoded()));
    } catch (IOException ex) {
      throw new CertificateException("Error while decoding X509 object");
    }
  }

  /**
   * Convert java.security.cert.X509Certificate objects to
   * org.bouncycastle.asn1.x509.X509CertificateStructure objects.
   *
   * @param cert X509Certificate object
   * @return X509CertificateStructure object
   * @throws CertificateEncodingException if the object cannot be encoded
   */
  static public X509CertificateStructure
      X509CertificateToX509CertificateStructure (X509Certificate cert)
      throws CertificateEncodingException {

    try {
        return new X509CertificateStructure
            ((ASN1Sequence)new ASN1InputStream
            (new ByteArrayInputStream(cert.getEncoded())).readObject());
    } catch (IOException ex) {
      throw new CertificateEncodingException("Error while encoding X509 object");
    }
  }

  /**
   * Convert org.bouncycastle.asn1.x509.SubjectPublicKeyInfo objects to
   * java.security.PublicKey objects
   *
   * @param subjectPKInfo SubjectPublicKeyInfo object
   * @return PublicKey object
   * @throws KeyException if the object cannot be decoded
   * @see PublicKeyToSubjectPublicKeyInfo
   */
  static public PublicKey SubjectPublicKeyInfoToPublicKey
      (SubjectPublicKeyInfo subjectPKInfo) throws KeyException {

    try {
      X509EncodedKeySpec xspec = new X509EncodedKeySpec
                                 (new DERBitString(subjectPKInfo).getBytes());
      AlgorithmIdentifier keyAlg = subjectPKInfo.getAlgorithmId ();
      return KeyFactory.getInstance(keyAlg.getObjectId().getId(), "BC")
                       .generatePublic(xspec);
    } catch (GeneralSecurityException ex) {
      throw new KeyException("Error while decoding public key");
    }
    
  }

  /**
   * Convert java.security.PublicKey objects to
   * org.bouncycastle.asn1.x509.SubjectPublicKeyInfo objects
   *
   * @param publicKey PublicKey object
   * @return SubjectPublicKeyInfo object
   * @throws InvalidKeyException if the object cannot be encoded
   * @see SubjectPublicKeyInfoToPublicKey
   */
  static public SubjectPublicKeyInfo PublicKeyToSubjectPublicKeyInfo
      (PublicKey publicKey) throws InvalidKeyException {

    try {
     return new SubjectPublicKeyInfo((ASN1Sequence) new ASN1InputStream(
                new ByteArrayInputStream(publicKey.getEncoded()))
                .readObject());
    } catch (IOException ex) {
      throw new InvalidKeyException("Error while encoding public key");
    }
  }

  /**
   * Converting Common Names to Strings
   * 
   * @param name Principal object
   * @return Array of string represent the common names
   */
  static public String[] getCommonNames(Principal name) {
    return getCommonNames(new X509Name(name.getName()));
  }

  /**
   * Converting Common Names to Strings
   *
   * @param name X509Name object
   * @return Array of string represent the common names
   */
  static public String[] getCommonNames(X509Name name) {
    Vector cns = name.getValues(X509Name.CN);
    String[] names = new String[cns.size()];
    for (int i = 0 ; i < cns.size() ; i++) {
      names[i] = (String)cns.get(i);
    }
    return names;
  }

  /**
   * Converting General Names to Byte String
   *
   * <PRE>
   * GeneralName ::= CHOICE {
   *   otherName             [0]   OtherName,
   *   rfc822Name            [1]   IA5String,
   *   dNSName               [2]   IA5String,
   *   x400Address           [3]   ORAddress,
   *   directoryName         [4]   Name,
   *   ediPartyName          [5]   EDIPartyName,
   *   uniformResourceIdentifier     [6]   IA5String,
   *   iPAddress             [7]   OCTET STRING,
   *   registeredID          [8]   OBJECT IDENTIFIER}
   *
   * OtherName ::= SEQUENCE {
   *   type-id  OBJECT IDENTIFIER,
   *   value    [0] EXPLICIT ANY DEFINED BY type-id }
   *
   * EDIPartyName ::= SEQUENCE {
   *   nameAssigner      [0]   DirectoryString OPTIONAL,
   *   partyName         [1]   DirectoryString }
   *
   * Name ::= CHOICE { RDNSequence }
   * </PRE>
   *
   * @param name General Name
   * @return Byte String
   * @throws IOException
   */
  static public byte[] getGeneralName(GeneralName name)
          throws IOException {
    switch(name.getTagNo()) {
      case GeneralName.otherName:
        return name.getName().getDERObject().getEncoded();
      case GeneralName.rfc822Name:
      case GeneralName.dNSName:
      case GeneralName.uniformResourceIdentifier:
        return ((DERIA5String)name.getName()).getOctets();
      case GeneralName.registeredID:
        return ((DERObjectIdentifier)name.getName()).getId().getBytes();
      case GeneralName.directoryName:
        return ((X509Name)name.getName()).toString().getBytes();
      case GeneralName.iPAddress:
        return ((DEROctetString)name.getName()).getOctets();
      default:
        throw null;
    }
  }

  /**
   * Create Random Nonces
   * 
   * @param len Length of nonces
   * @return Byte array
   */
  static public byte[] createNonce(int len) {
    byte[] nonce = new byte[len];
    new Random().nextBytes(nonce);
    return nonce;
  }

  /**
   * Combine encryption algorithm and digest algorithm
   *
   * @param encId encryption id
   * @param digestId digest id
   * @return DERObjectIdentifer of the algorithm
   */
  static public DERObjectIdentifier createSigId(DERObjectIdentifier encId,
                                                DERObjectIdentifier digestId) {

    if (encId.equals(PKCSObjectIdentifiers.rsaEncryption)) {
      if (digestId.equals(PKCSObjectIdentifiers.md2))
        return PKCSObjectIdentifiers.md2WithRSAEncryption;
      if (digestId.equals(PKCSObjectIdentifiers.md4))
        return PKCSObjectIdentifiers.md4WithRSAEncryption;
      if (digestId.equals(PKCSObjectIdentifiers.md5))
        return PKCSObjectIdentifiers.md5WithRSAEncryption;
      if (digestId.equals(OIWObjectIdentifiers.idSHA1))
        return PKCSObjectIdentifiers.sha1WithRSAEncryption;
      if (digestId.equals(NISTObjectIdentifiers.id_sha224))
        return PKCSObjectIdentifiers.sha224WithRSAEncryption;
      if (digestId.equals(NISTObjectIdentifiers.id_sha256))
        return PKCSObjectIdentifiers.sha256WithRSAEncryption;
      if (digestId.equals(NISTObjectIdentifiers.id_sha384))
        return PKCSObjectIdentifiers.sha384WithRSAEncryption;
      if (digestId.equals(NISTObjectIdentifiers.id_sha512))
        return PKCSObjectIdentifiers.sha512WithRSAEncryption;
      return null;
    }
    return null;
  }
}