package org.netserv.ipauth.util;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.security.KeyPair;
import java.security.SecureRandom;
import java.security.cert.X509Certificate;
import java.util.Vector;
import org.bouncycastle.asn1.x509.X509CertificateStructure;
import org.bouncycastle.openssl.PEMReader;
import org.bouncycastle.openssl.PEMWriter;
import org.bouncycastle.openssl.PasswordFinder;

/**
 * This helper provided utilities for reading or writing cryptographic objects
 * from files.
 *
 * @author Chia-che Tsai
 */

public class FileHelper {
  
  /**
   * Read Certificate Chain from PEM files
   *
   * @param pemFile the PEM file
   * @return Array of X509 certificates
   */
  public static X509Certificate[] readCertificates(String pemFile) {
    Vector<X509Certificate> certs = new Vector<X509Certificate>();
    try {
      PEMReader reader = new PEMReader(new FileReader(pemFile));
      Object o;
      while((o = reader.readObject()) != null) {
        if (o instanceof X509Certificate)
          certs.add((X509Certificate)o);
      }

    } catch (IOException ex) { return null; }
    return certs.toArray(new X509Certificate[certs.size()]);
  }

  /**
   *
   * @param pemFile
   * @return
   */
  public static X509Certificate readSingleCertificate(String pemFile) {
    try {
      PEMReader reader = new PEMReader(new FileReader(pemFile));
      Object o = reader.readObject();
      if (o instanceof X509Certificate)
        return(X509Certificate)o;
      else
        return null;
    } catch (IOException ex) { return null; }
  }

  /**
   *
   * @param pemFile
   * @return
   */
  public static X509CertificateStructure[]
          readCertificateStructures(String pemFile) {
    Vector<X509CertificateStructure> certs =
            new Vector<X509CertificateStructure>();
    try {
      PEMReader reader = new PEMReader(new FileReader(pemFile));
      Object o;
      while((o = reader.readObject()) != null) {
        if (o instanceof X509Certificate) {
          try {
            certs.add(CryptoHelper.X509CertificateToX509CertificateStructure
                      ((X509Certificate)o));
          } catch (Exception ex) {}
        }
      }

    } catch (IOException ex) { return null; }
    return certs.toArray(new X509CertificateStructure[certs.size()]);
  }

  /**
   *
   * @param pemFile
   * @return
   */
  public static X509CertificateStructure
          readSingleCertificateStructure(String pemFile) {
    try {
      PEMReader reader = new PEMReader(new FileReader(pemFile));
      Object o = reader.readObject();
      if (o instanceof X509Certificate) {
        try {
          return CryptoHelper.X509CertificateToX509CertificateStructure
                        ((X509Certificate)o);
        } catch (Exception ex) {}
      }
    } catch (IOException ex) {}
    return null;
  }

  /**
   *
   * @param pemFile
   * @param password
   * @return
   */
  public static KeyPair readKeyPair(String pemFile, String password) {
    KeyPair key = null;

    PasswordFinder pFinder = null;
    if (password != null) {
      final char[] myPass = password.toCharArray();
      pFinder = new PasswordFinder() {
        public char[] getPassword() {
          return myPass;
        }
      };
    }
    else {
      pFinder = new PasswordFinder() {
        public char[] getPassword() {
          return passwordHandler();
        }
      };
    }
    try {
      PEMReader reader = new PEMReader(new FileReader(pemFile), pFinder);
      Object o = reader.readObject();
      if (o != null && o instanceof KeyPair) {
        key = (KeyPair)o;
      }

    } catch(IOException ex) {
    	ex.printStackTrace();
    }

    return key;
  }

  /**
   *
   * @param certs
   * @param pemFile
   * @throws IOException
   */
  public static void writeCertificates(X509Certificate[] certs, String pemFile)
          throws IOException {
    PEMWriter writer = new PEMWriter(new FileWriter(pemFile));
    for (X509Certificate cert: certs) {
      writer.writeObject(cert);
    }
    writer.close();
  }

  /**
   *
   * @param cert
   * @param pemFile
   * @throws IOException
   */
  public static void writeSingleCertificate(X509Certificate cert, String pemFile)
          throws IOException {
    PEMWriter writer = new PEMWriter(new FileWriter(pemFile));
    writer.writeObject(cert);
    writer.close();
  }

  /**
   *
   * @param certs
   * @param pemFile
   * @throws IOException
   */
  public static void writeCertificateStructures(X509CertificateStructure[] certs, String pemFile)
          throws IOException {
    PEMWriter writer = new PEMWriter(new FileWriter(pemFile));
    for (X509CertificateStructure cert: certs) {
      writer.writeObject(cert);
    }
    writer.close();
  }

  /**
   *
   * @param cert
   * @param pemFile
   * @throws IOException
   */
  public static void writeSingleCertificateStructure(X509CertificateStructure cert, String pemFile)
          throws IOException {
    PEMWriter writer = new PEMWriter(new FileWriter(pemFile));
    writer.writeObject(cert);
    writer.close();
  }

  /**
   *
   * @param key
   * @param pemFile
   * @throws IOException
   */
  public static void writeKeyPair(KeyPair key, String pemFile)
          throws IOException {
    PEMWriter writer = new PEMWriter(new FileWriter(pemFile));
    writer.writeObject(key);
    writer.close();
  }

  /**
   *
   * @param pemFile
   * @param password
   * @return
   */
  public static void writeKeyPair(KeyPair key, String pemFile, String password) {
    char[] passwordChars;
    if (password != null) {
      passwordChars = password.toCharArray();
    }
    else {
      passwordChars = passwordHandler();
    }
    try {
      PEMWriter writer = new PEMWriter(new FileWriter(pemFile));
      writer.writeObject(key, "DESEDE", passwordChars, new SecureRandom());
    } catch(IOException ex) {}

  }

  protected final static char[] passwordHandler() {
    BufferedReader input = new BufferedReader(new InputStreamReader(System.in));

    System.err.print("Enter Password: ");

    String myPass = null;
    try {
      myPass = input.readLine();
    } catch (IOException ex) {
      return null;
    }
    return myPass.toCharArray();
  }
}
