#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/bn.h>
#include <openssl/rsa.h>
#include <openssl/des.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/rand.h>
#include <openssl/md5.h>

#include "ipauth.h"
#include "crypto.h"

#define BUFFSIZE 4096

BIO *bio_err;

struct ipauth_crypto_config_t crypto_config = {
    .protect_password   = NULL,
    .protect_method     = NULL,
    .random_device      = "/dev/urandom",
};

void CHECK_AND_SET_bio_err(void)
{
	if (bio_err)
		return;
	if ((bio_err = BIO_new(BIO_s_file())) != NULL)
		BIO_set_fp(bio_err, stderr, BIO_NOCLOSE | BIO_FP_TEXT);
}

struct key_suite *genkey(unsigned char method, unsigned short ksize)
{
    struct key_suite *key_suite;

    CHECK_AND_SET_bio_err();

    switch (method) {
        case METHOD_RSA_PKCS1_PADDING:
        case METHOD_RSA_PKCS1_OAEP_PADDING:
        case METHOD_RSA_SSLV23_PADDING:
        case METHOD_RSA_NO_PADDING: {

	        unsigned long f4 = RSA_F4;
	
	        BIGNUM *bn = BN_new();
	        RSA *rsa = RSA_new();

	        if(!bn || !rsa) goto rsa_err;

	        ERR_load_crypto_strings();
	
	        if (!RAND_load_file(crypto_config.random_device, ksize) 
                        && !RAND_status()) {
		        BIO_printf(bio_err, "cannot get random data\n");
	        }
	
	        if(!BN_set_word(bn, f4) || !RSA_generate_key_ex(rsa, ksize, bn, NULL))
		        goto rsa_err;

	        if (bn) BN_free(bn);

            key_suite = malloc(sizeof(struct key_suite));
            key_suite->method = method;
            key_suite->ksize = RSA_size(rsa);
            key_suite->key = (void *)rsa;
            break;

        rsa_err:
	        if (bn) BN_free(bn);
	        if (rsa) RSA_free(rsa);
        }

        default:
            return NULL;
    }

    return key_suite;
}

void write_key_to_bio(struct key_suite *key, BIO *bio_out, char public)
{
	const EVP_CIPHER *enc = NULL;
	
	CHECK_AND_SET_bio_err();

    if (public) {
        switch (key->method) {
            case METHOD_RSA_PKCS1_PADDING:
                BIO_puts(bio_out, "using RSA_PKCS1_PADDING\n");
                goto write_public;

            case METHOD_RSA_PKCS1_OAEP_PADDING:
                BIO_puts(bio_out, "using RSA_PKCS1_OAEP_PADDING\n");
                goto write_public;

            case METHOD_RSA_SSLV23_PADDING:
                BIO_puts(bio_out, "using RSA_SSLV23_PADDING\n");
                goto write_public;

            case METHOD_RSA_NO_PADDING:
                BIO_puts(bio_out, "using RSA_NO_PADDING\n");
                goto write_public;
            {
            write_public:
                    if (!PEM_write_bio_RSA_PUBKEY(bio_out, (RSA *)key->key))
			        goto err;
                
                break;
            }
            default:
                goto err;
        }
    }
    else {
		if (crypto_config.protect_method == NULL || 
            crypto_config.protect_password == NULL)
            enc = NULL;
#ifndef OPENSSL_NO_DES
		else if (strcmp(crypto_config.protect_method, "des") == 0)
			enc = EVP_des_cbc();
		else if (strcmp(crypto_config.protect_method, "des3") == 0)
			enc = EVP_des_ede3_cbc();
#endif
#ifndef OPENSSL_NO_IDEA
		else if (strcmp(crypto_config.protect_method, "idea") == 0)
			enc = EVP_idea_cbc();
#endif
#ifndef OPENSSL_NO_SEED
		else if (strcmp(crypto_config.protect_method, "seed") == 0)
			enc = EVP_seed_cbc();
#endif
#ifndef OPENSSL_NO_AES
		else if (strcmp(crypto_config.protect_method, "aes128") == 0)
			enc = EVP_aes_128_cbc();
		else if (strcmp(crypto_config.protect_method, "aes192") == 0)
			enc = EVP_aes_192_cbc();
		else if (strcmp(crypto_config.protect_method, "aes256") == 0)
			enc = EVP_aes_256_cbc();
#endif
#ifndef OPENSSL_NO_CAMELLIA
		else if (strcmp(crypto_config.protect_method, "camellia128") == 0)
			enc = EVP_camellia_128_cbc();
		else if (strcmp(crypto_config.protect_method, "camellia192") == 0)
			enc = EVP_camellia_192_cbc();
		else if (strcmp(crypto_config.protect_method, "camellia256") == 0)
			enc = EVP_camellia_256_cbc();
#endif
		else
			goto err;
	        
        switch (key->method) {
            case METHOD_RSA_PKCS1_PADDING:
                BIO_puts(bio_out, "using RSA_PKCS1_PADDING\n");
                goto write_private;

            case METHOD_RSA_PKCS1_OAEP_PADDING:
                BIO_puts(bio_out, "using RSA_PKCS1_OAEP_PADDING\n");
                goto write_private;

            case METHOD_RSA_SSLV23_PADDING:
                BIO_puts(bio_out, "using RSA_SSLV23_PADDING\n");
                goto write_private;

            case METHOD_RSA_NO_PADDING:
                BIO_puts(bio_out, "using RSA_NO_PADDING\n");
                goto write_private;
             
            {
            write_private:
		        if (!PEM_write_bio_RSAPrivateKey(bio_out, (RSA *)key->key, 
                            enc, NULL, 0, NULL, crypto_config.protect_password))
			        goto err;
                
                break;
            }
            default:
                goto err;
        }
	}

err:
	return;
}

struct key_suite *read_key_from_bio(BIO *bio_in, char public)
{
	struct key_suite *key = NULL;
	char *buff = malloc(BUFFSIZE), *tok;
    unsigned char method = 0;

	CHECK_AND_SET_bio_err();
	OpenSSL_add_all_ciphers();

    BIO_gets(bio_in, buff, BUFFSIZE);
    tok = strtok(buff, " \t\n");
    while (tok != NULL) {
        if (strcmp(tok, "using") == 0);
        else if (strcmp(tok, "RSA_PKCS1_PADDING") == 0) {
            method = METHOD_RSA_PKCS1_PADDING;
            
            free(buff);
            break;
        }
        else if (strcmp(tok, "RSA_PKCS1_OAEP_PADDING") == 0) {
            method = METHOD_RSA_PKCS1_OAEP_PADDING;
            
            free(buff);
            break;
        }
        else if (strcmp(tok, "RSA_SSLV23_PADDING") == 0) {
            method = METHOD_RSA_SSLV23_PADDING;
            
            free(buff);
            break;
        }
        else if (strcmp(buff, "using RSA_PKCS1_PADDING") == 0) {
            method = METHOD_RSA_PKCS1_PADDING;
            
            free(buff);
            break;
        }
        else {
            free(buff);
            return NULL;
        }
        tok = strtok(NULL, " \t\n");
    }

    if (method == 0)
        return NULL;

    switch (method) {
        case METHOD_RSA_PKCS1_PADDING:
        case METHOD_RSA_PKCS1_OAEP_PADDING:
        case METHOD_RSA_SSLV23_PADDING:
        case METHOD_RSA_NO_PADDING: {
	        RSA *rsa;

            if (public) {
		        rsa = PEM_read_bio_RSA_PUBKEY(bio_in, NULL, NULL,
                            crypto_config.protect_password);
	        }
        	else {
		        rsa = PEM_read_bio_RSAPrivateKey(bio_in, NULL, NULL, 
                            crypto_config.protect_password);
	        }
 
	        if (!rsa) {
		        ERR_print_errors(bio_err);	
		        goto err;
	        }

            key = malloc(sizeof(struct key_suite));
            key->method = method;
            key->ksize = RSA_size(rsa);
            key->key = (void *)rsa;

            break;
	    }
        default:
            goto err;
    }
err:
	return key;
}


void write_key_to_file(struct key_suite *key, char *out_file, char public)
{
	BIO *bio_out;
	
	CHECK_AND_SET_bio_err();

	if ((bio_out = BIO_new(BIO_s_file())) == NULL) {
		BIO_printf(bio_err, "unable to create BIO for output\n");
		goto err;
	}

	if (out_file == NULL) {
		BIO_set_fp(bio_out, stdout, BIO_NOCLOSE);
	}
	else {
		if (BIO_write_filename(bio_out, out_file) <= 0) {
			perror(out_file);
			goto err;
		}
	}

	write_key_to_bio(key, bio_out, public);

err:
	if (bio_out) BIO_free_all(bio_out);
	return;
}

struct key_suite *read_key_from_file(char *in_file, char public)
{
	BIO *bio_in;
	struct key_suite *key = NULL;
	
	CHECK_AND_SET_bio_err();

	if ((bio_in = BIO_new(BIO_s_file())) == NULL) {
		BIO_printf(bio_err, "unable to create BIO for input\n");
		goto err;
	}

	if (in_file == NULL) {
		BIO_set_fp(bio_in, stdin, BIO_NOCLOSE);
	}
	else {
		if (BIO_read_filename(bio_in, in_file) <= 0) {
			perror(in_file);
			goto err;
		}
	}

	key = read_key_from_bio(bio_in, public);
	if (!key) {
		ERR_print_errors(bio_err);	
		goto err;
	}

err:
	if (bio_in) BIO_free_all(bio_in);
	return key;
}

int do_encryption(int flen, void *to, void *from,
                        struct key_suite *key_suite) {
	int ret = 0;

    switch(key_suite->method) {
        case METHOD_RSA_PKCS1_PADDING:
            ret = RSA_public_encrypt(flen, 
                        (unsigned char *)from, (unsigned char *)to, 
                        (RSA *)key_suite->key, RSA_PKCS1_PADDING);
            break;

        case METHOD_RSA_PKCS1_OAEP_PADDING:
            ret = RSA_public_encrypt(flen, 
                        (unsigned char *)from, (unsigned char *)to, 
                        (RSA *)key_suite->key, RSA_PKCS1_OAEP_PADDING);
            break;

        case METHOD_RSA_SSLV23_PADDING:
            ret = RSA_public_encrypt(flen, 
                        (unsigned char *)from, (unsigned char *)to, 
                        (RSA *)key_suite->key, RSA_SSLV23_PADDING);
            break;

        case METHOD_RSA_NO_PADDING: 
            ret = RSA_public_encrypt(flen, 
                        (unsigned char *)from, (unsigned char *)to, 
                        (RSA *)key_suite->key, RSA_NO_PADDING);
            break;

        default:
            return 0;
    }

	return ret;
}

int do_decryption(int flen, void *to, void *from, 
                        struct key_suite *key_suite) {
	int ret = 0;

    switch(key_suite->method) {
        case METHOD_RSA_PKCS1_PADDING:
            ret = RSA_private_decrypt(flen, 
                        (unsigned char *)from, (unsigned char *)to,
                        (RSA *)key_suite->key, RSA_PKCS1_PADDING);
            break;

        case METHOD_RSA_PKCS1_OAEP_PADDING:
            ret = RSA_private_decrypt(flen,
                        (unsigned char *)from, (unsigned char *)to,
                        (RSA *)key_suite->key, RSA_PKCS1_OAEP_PADDING);
            break;

        case METHOD_RSA_SSLV23_PADDING:
            ret = RSA_private_decrypt(flen, 
                        (unsigned char *)from, (unsigned char *)to, 
                        (RSA *)key_suite->key, RSA_SSLV23_PADDING);
            break;

        case METHOD_RSA_NO_PADDING: 
            ret = RSA_private_decrypt(flen, 
                        (unsigned char *)from, (unsigned char *)to, 
                        (RSA *)key_suite->key, RSA_NO_PADDING);
            break;

        default:
            return 0;
    }
	
	if (ret == -1) {
        return 0;
    }

	return ret;
}

int i2d_key(struct key_suite *key, unsigned char *buff, char public) {
    unsigned char *bp = buff;

    if (key == NULL)
        return 0;
 
    switch (key->method) {
        case METHOD_RSA_PKCS1_PADDING:
        case METHOD_RSA_PKCS1_OAEP_PADDING:
        case METHOD_RSA_SSLV23_PADDING:
        case METHOD_RSA_NO_PADDING: {
	        if (public) {
                return i2d_RSA_PUBKEY((RSA *)key->key, &bp);
            }
            else {
                return i2d_RSAPrivateKey((RSA *)key->key, &bp);
            }
        }
        default:
            return 0;
    }
}

struct key_suite *d2i_key(unsigned char method, 
            unsigned char *buff, int len, char public) {
    const unsigned char *bp = buff;
 
    switch (method) {
        case METHOD_RSA_PKCS1_PADDING:
        case METHOD_RSA_PKCS1_OAEP_PADDING:
        case METHOD_RSA_SSLV23_PADDING:
        case METHOD_RSA_NO_PADDING: {
            struct key_suite *key = malloc(sizeof(struct key_suite));
            key->method = method;

	        if (public) {
                key->key = d2i_RSA_PUBKEY(NULL, &bp, len);
            }
            else {
                key->key = d2i_RSAPrivateKey(NULL, &bp, len);
            }
            key->ksize = RSA_size(key->key);

            return key;
        }
        default:
            return NULL;
    }
}


unsigned char key_id(struct key_suite *key) {
    unsigned char md5[MD5_DIGEST_LENGTH];
    unsigned char buff[BUFFSIZE];
    int len = i2d_key(key, buff, 1);
    
    MD5(buff, len, md5);
    return md5[0];
}

int cert_hash(X509 *cert, void *to, int size) {
    unsigned char md5[MD5_DIGEST_LENGTH];
    unsigned char buff[BUFFSIZE];
    const unsigned char *bp = buff;
    int len = i2d_X509(cert, &bp);
    
    MD5(buff, len, md5); 
    memcpy(to, buff, size);

    return size;
}

X509_REQ *create_x509_req(unsigned char *cn, struct key_suite *key)
{
    EVP_PKEY *pkey = EVP_PKEY_new();
    X509_REQ *req = X509_REQ_new();
    X509_NAME *n = X509_NAME_new();

	if (!X509_NAME_add_entry_by_txt(
                    n, "CN", MBSTRING_UTF8, cn, -1, -1, 0))
			goto err;
    
    X509_REQ_set_subject_name(req, n);

    switch (key->method) {
        case METHOD_RSA_PKCS1_PADDING:
        case METHOD_RSA_PKCS1_OAEP_PADDING:
        case METHOD_RSA_SSLV23_PADDING:
        case METHOD_RSA_NO_PADDING: {
            EVP_PKEY_set1_RSA(pkey, (RSA *)key->key);
            break;
        }
        default:
            goto err;
   }

    X509_REQ_set_pubkey(req, pkey);

    EVP_PKEY_free(pkey);
    return req;

err:
    if (pkey) EVP_PKEY_free(pkey);
    if (req) X509_REQ_free(req);
    if (n) X509_NAME_free(n);
    return NULL;
}

X509_REQ *sign_x509_req(X509_REQ *req, struct key_suite *key)
{
    EVP_PKEY *pkey = EVP_PKEY_new();

    switch (key->method) {
        case METHOD_RSA_PKCS1_PADDING:
        case METHOD_RSA_PKCS1_OAEP_PADDING:
        case METHOD_RSA_SSLV23_PADDING:
        case METHOD_RSA_NO_PADDING: {
            EVP_PKEY_set1_RSA(pkey, (RSA *)key->key);
            break;
        }
        default:
            goto err;
   }

    X509_REQ_sign(req, pkey, EVP_sha1());

    EVP_PKEY_free(pkey);
    return req;

err:
    if (pkey) EVP_PKEY_free(pkey);
    return NULL;
}

