#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <openssl/x509.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "crypto.h"
#include "ipauth.h"

#define BUFFSIZE 1024

int do_ipauth_option_encapsulate(struct ipauth_option *option, 
                                 uint8_t *to, int len, struct key_suite *key)
{
    int wbyte = 0;

    if (option->enc && key) {
        unsigned char *buff = malloc(BUFFSIZE);
        uint8_t keyid = key_id(key);
        int len = 0;
        
        if (key != NULL) {
            len = do_encryption(
                    sizeof(struct ipauth_encrypted) + option->enc->extlen, 
                    buff, option->enc, key); 
            if (len <= 0) {
                fprintf(stderr, "cannot do encryption\n");
                ERR_print_errors_fp(stderr);
            }
        }
        
        if (len > 0 && len < 256) {
          
            to[wbyte++] = keyid;
            to[wbyte++] = len;

            memcpy(to + wbyte, buff, len);
        
            fprintf(stderr, "IPAUTH INFO encrypted, plain len = %d(%lx,%lx)\n", 
                    sizeof(struct ipauth_encrypted) + option->enc->extlen,
                    option->enc->xid, option->enc->yaddr);
            fprintf(stderr, "IPAUTH ENCRYPTION IMPLIED, len = %d\n", len);
        
            wbyte += len;
        }
        else {
            to[wbyte++] = 0;
            to[wbyte++] = 0;
        }
        free(buff);
    }
    else {
        to[wbyte++] = 0;
        to[wbyte++] = 0;
    }

    if (option->subnum && option->sub) {
        int i = 0;

        for (i = 0 ; i < option->subnum ; i++) {
            struct ipauth_suboption *sub = &option->sub[i];
            if (!sub) break;
            int size = sizeof(struct ipauth_suboption) + sub->optlen;

            memcpy(to + wbyte, sub, size);
            wbyte += size;
        }
    }

    fprintf(stderr, "IPAUTH CREATED, total_len = %d\n", wbyte);

    return wbyte;
}

int do_ipauth_option_decapsulate(struct ipauth_option *option, 
                                 uint8_t *from, int len, struct key_suite *key)
{
    int rbyte = 0;
    
    uint8_t keyid;
    uint8_t enclen;

    memcpy(&keyid, from + rbyte, sizeof(keyid));
    rbyte += sizeof(keyid);

    memcpy(&enclen, from + rbyte, sizeof(enclen));
    rbyte += sizeof(enclen);

    if (enclen && key && keyid == key_id(key)) {
        unsigned char *buff = malloc(BUFFSIZE);
        uint8_t len = do_decryption(enclen, buff, from + rbyte, key); 
        struct ipauth_encrypted *enc = malloc(len);
            
        memcpy(enc, buff, len);
        enc->extlen = len;
        option->enc = enc;
        rbyte += enclen;
    }
    else {
        option->enc = NULL;
        rbyte += enclen;
    }

    if (rbyte < len) {
        struct ipauth_suboption *ps = NULL;
        option->sub = malloc(len - rbyte);
        memcpy(option->sub, from + rbyte, len - rbyte);
        option->subnum = 0;

        ps = option->sub;
        while(rbyte < len) {
            int size = sizeof(struct ipauth_suboption) + ps->optlen;
            rbyte += size;
            ps += size;
            option->subnum++;
        }
    }

    return rbyte;
}

int make_provide_key(struct ipauth_provide_key **sub, struct key_suite *key) {
    struct ipauth_provide_key *sp = NULL;
    unsigned char *buff = malloc(BUFFSIZE);
    int len = i2d_key(key, buff, 1);
    int total_len = sizeof(struct ipauth_provide_key) + len;

    sp = malloc(total_len);

    sp->method = key->method;
    sp->keylen = len;
    memcpy(&sp->key, buff, len);

    *sub = sp;

    free(buff);
    return total_len;
}

struct key_suite *get_provide_key(struct ipauth_provide_key *sub) {
    return d2i_key(sub->method, sub->key, sub->keylen, 1); 
}

int make_cert_request(struct ipauth_cert_request **sub, X509_REQ *req) {
    // Might not necessary
    return 0;
}

X509_REQ *get_cert_request(struct ipauth_cert_request *sub) {
    // Might not necessary
    return NULL;
}

int make_cert_deploy(struct ipauth_cert_deploy **sub, X509 *cert) {
    struct ipauth_cert_deploy *sp = NULL;
    unsigned char *buff = malloc(BUFFSIZE);
    unsigned char *bp = buff;
    int len = i2d_X509(cert, &bp);
    int total_len = sizeof(struct ipauth_cert_deploy);

    sp = malloc(total_len);

    sp->cerlen = len;
    cert_hash(cert, &sp->cerhash, sizeof(uint64_t));

    fprintf(stderr, "cerlen = %X, cerhash = %llX\n", sp->cerlen, sp->cerhash);

    *sub = sp;
    free(buff);
    return total_len;
}

int check_cert_deploy(struct ipauth_cert_deploy *sub, X509 *cert) {
    uint64_t hash;
    cert_hash(cert, &hash, sizeof(uint64_t));
    
    fprintf(stderr, "check %llx with %llx\n", hash, sub->cerhash);

    if (hash == sub->cerhash)
        return 1;
    else return 0;
}

int make_ipauth_option(uint32_t xid, uint32_t ip, struct key_suite *key, 
            unsigned char subcode, void *obj, unsigned char *to, int len) {

    struct ipauth_option *option = malloc(sizeof(struct ipauth_option));
    struct ipauth_suboption *subopt =  NULL;
    void *new_obj = NULL;
    int got = 0;

    fprintf(stderr, "subcode = %d\n", subcode);

    switch(subcode) {
    case IPAUTH_PROVIDE_KEY:
        got = make_provide_key((struct ipauth_provide_key **)&new_obj,
                               (struct key_suite *)obj);
        break;
    case IPAUTH_CERT_REQUEST:
        break;
    case IPAUTH_CERT_DEPLOY:
        got = make_cert_deploy((struct ipauth_cert_deploy **)&new_obj,
                               (X509 *)obj);
        break;
    default:
        break;
    }

    fprintf(stderr, "got = %d\n", got);

    if (new_obj && got) {
        fprintf(stderr, "Add a new ipauth suboption\n");

        subopt =  malloc(sizeof(struct ipauth_suboption) + got);
        subopt->code = subcode;
        subopt->optlen = got;
        memcpy(&subopt->opt, new_obj, got);
        free(new_obj);

        option->subnum = 1;
        option->sub = subopt;
    }
    else {
        option->subnum = 0;
        option->sub = NULL;
    }

    option->enc = malloc(sizeof(struct ipauth_encrypted));
    option->enc->xid = xid;
    option->enc->yaddr = ip;
    option->enc->extlen = 0;

    fprintf(stderr, "Building the IPAUTH option\n");

    return do_ipauth_option_encapsulate(option, to, len, key);
}

int read_ipauth_option(uint32_t *xid, uint32_t *ip, struct key_suite *key, 
            unsigned char *subcode, void **obj, unsigned char *from, int len) {
    struct ipauth_option *option = malloc(sizeof(struct ipauth_option));
    int got = do_ipauth_option_decapsulate(option, from ,len, key);


    if (option->enc) {
        *xid = option->enc->xid;
        *ip = option->enc->yaddr;
    }
    else {
        *xid = 0;
        *ip = 0;
    }

    if (option->subnum == 0) {
        *subcode = 0;
        *obj = NULL;
    }
    else {
        switch(option->sub[0].code) {
        case IPAUTH_PROVIDE_KEY:
            *subcode = option->sub[0].code;
            *obj = (void *)get_provide_key((struct ipauth_provide_key *)
                                           option->sub[0].opt);
        break;
        case IPAUTH_CERT_REQUEST:
        break;
        case IPAUTH_CERT_DEPLOY:
            *subcode = option->sub[0].code;
            *obj = option->sub[0].opt;
        break;
        default:
        break;
        }
    }

    return got;
}

X509 *ask_for_cert(uint32_t ip, struct key_suite *key) {
        unsigned char *buff = malloc(BUFFSIZE);
        const unsigned char *bp = buff;
        char *dn = malloc(80);
        int wbyte = 0, rbyte = 0, i;
        unsigned char *size, *len;
        int sockfd;
        struct sockaddr_in serv_addr;
        X509 *x509;

        sprintf(dn, "CN=%d.%d.%d.%d", ((uint8_t*)&ip)[0], ((uint8_t*)&ip)[1], 
                    ((uint8_t*)&ip)[2], ((uint8_t*)&ip)[3]);

        buff[wbyte++] = 0x30;
        size = &buff[wbyte++];
        buff[wbyte++] = 0x0c;
        buff[wbyte++] = strlen(dn);
        memcpy(buff + wbyte, dn , strlen(dn));
        wbyte += strlen(dn);
        wbyte += i2d_key(key, buff + wbyte, 1); 
        buff[wbyte++] = 0x0a;
        *size = wbyte - 3;

        sockfd = socket(AF_INET, SOCK_STREAM, 0);
        serv_addr.sin_family = AF_INET;
        serv_addr.sin_port = htons(100);
        serv_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
        if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
            free(buff);
            free(dn);
            return NULL;
        }

        fprintf(stderr, "IPAUTH: send out request:");
        for (i = 0 ; i < wbyte ; i++) 
            fprintf(stderr, " %02X", buff[i]);
        fprintf(stderr, "\n");

        send(sockfd, buff, wbyte, 0);
        fprintf(stderr, "IPAUTH: send out request to CMP client\n");
        rbyte = recv(sockfd, buff, BUFFSIZE, 0);
        fprintf(stderr, "IPAUTH: receive response (%d bytes) from CMP client\n", rbyte);
        close(sockfd);

        fprintf(stderr, "IPAUTH: send out request:");
        for (i = 0 ; i < rbyte ; i++) 
            fprintf(stderr, " %02X", buff[i]);
        fprintf(stderr, "\n");

        {
            FILE *out = fopen("testcert", "w");
            fwrite(buff, 1, rbyte, out);
            fclose(out);
        }

        x509 = d2i_X509(NULL, &bp, rbyte);
        if (x509 == NULL)
            fprintf(stderr, "IPAUTH: cannot parse as X509 cert\n");


        free(buff);
        free(dn);
        return x509;
}

